#!/usr/bin/env python
# Copyright (C) 2018 Sorin Sbarnea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import sys

from jenkins_jobs.builder import JenkinsManager
import jenkins_jobs.cli.subcommand.base as base
import jenkins_jobs.utils as utils


def list_duplicates(seq):
    seen = set()
    return set(x for x in seq if x in seen or seen.add(x))


class ListSubCommand(base.JobsSubCommand):
    def parse_args(self, subparser):
        list = subparser.add_parser("list", help="List jobs")

        self.parse_option_recursive_exclude(list)

        list.add_argument("names", help="name(s) of job(s)", nargs="*", default=None)
        list.add_argument(
            "-p", "--path", default=None, help="path to YAML file or directory"
        )

    def execute(self, options, jjb_config):
        jobs = self.get_jobs(jjb_config, options.path, options.names)

        logging.info("Matching jobs: %d", len(jobs))
        stdout = utils.wrap_stream(sys.stdout)

        for job in jobs:
            stdout.write((job + "\n").encode("utf-8"))

    def get_jobs(self, jjb_config, path_list, glob_list):
        if path_list:
            roots = self.load_roots(jjb_config, path_list)
            jobs = base.filter_matching(roots.generate_jobs(), glob_list)
            job_names = [j.name for j in jobs]
        else:
            jenkins = JenkinsManager(jjb_config)
            job_names = [
                j["fullname"]
                for j in jenkins.get_jobs()
                if not glob_list or base.matches(j["fullname"], glob_list)
            ]

        job_names = sorted(job_names)
        for duplicate in list_duplicates(job_names):
            logging.warning("Found duplicate job name '%s', likely bug.", duplicate)

        logging.debug("Builder.get_jobs: returning %r", job_names)

        return job_names



# Copyright 2015 David Caro
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
from multiprocessing import cpu_count

from jenkins_jobs.parallel import concurrent


def test_parallel_correct_order():
    expected = list(range(10, 20))

    @concurrent
    def parallel_test(num_base, num_extra):
        return num_base + num_extra

    parallel_args = [{"num_extra": num} for num in range(10)]
    result = parallel_test(10, concurrent=parallel_args)
    assert result == expected


def test_parallel_time_less_than_serial():
    @concurrent
    def wait(secs):
        time.sleep(secs)

    before = time.time()
    # ten threads to make it as fast as possible
    wait(concurrent=[{"secs": 1} for _ in range(10)], n_workers=10)
    after = time.time()
    assert after - before < 5


def test_parallel_single_thread():
    expected = list(range(10, 20))

    @concurrent
    def parallel_test(num_base, num_extra):
        return num_base + num_extra

    parallel_args = [{"num_extra": num} for num in range(10)]
    result = parallel_test(10, concurrent=parallel_args, n_workers=1)
    result == expected


def test_use_auto_detect_cores(mocker):
    mock = mocker.patch("jenkins_jobs.parallel.cpu_count", wraps=cpu_count)

    @concurrent
    def parallel_test():
        return True

    result = parallel_test(concurrent=[{} for _ in range(10)], n_workers=0)
    assert result == [True for _ in range(10)]
    mock.assert_called_once_with()



#!/usr/bin/env python
# Copyright (C) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# functions that don't fit in well elsewhere

import codecs
import fnmatch
import locale
import os.path
from six.moves import input


def wrap_stream(stream, encoding="utf-8"):

    try:
        stream_enc = stream.encoding
    except AttributeError:
        stream_enc = locale.getpreferredencoding()

    if hasattr(stream, "buffer"):
        stream = stream.buffer

    if str(stream_enc).lower() == str(encoding).lower():
        return stream

    return codecs.EncodedFile(stream, encoding, stream_enc)


def recurse_path(root, excludes=None):
    if excludes is None:
        excludes = []

    basepath = os.path.realpath(root)
    pathlist = [basepath]

    patterns = [e for e in excludes if os.path.sep not in e]
    absolute = [e for e in excludes if os.path.isabs(e)]
    relative = [e for e in excludes if os.path.sep in e and not os.path.isabs(e)]
    for root, dirs, files in os.walk(basepath, topdown=True):
        # sort in-place to ensure dirnames are visited in alphabetical order
        # a predictable order makes it easier to use the retain_anchors option
        dirs.sort()
        dirs[:] = [
            d
            for d in dirs
            if not any([fnmatch.fnmatch(d, pattern) for pattern in patterns])
            if not any(
                [
                    fnmatch.fnmatch(os.path.abspath(os.path.join(root, d)), path)
                    for path in absolute
                ]
            )
            if not any(
                [
                    fnmatch.fnmatch(os.path.relpath(os.path.join(root, d)), path)
                    for path in relative
                ]
            )
        ]
        pathlist.extend([os.path.join(root, path) for path in dirs])

    return pathlist


def confirm(question):
    answer = input("%s (Y/N): " % question).upper().strip()
    return answer == "Y"



#!/usr/bin/env python
# Copyright (C) 2015 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage interpolation of JJB variables into template strings.

import _string
import logging
import re
from string import Formatter

from jinja2 import Undefined
from jinja2.exceptions import UndefinedError

from .errors import JenkinsJobsException

logger = logging.getLogger(__name__)


class CustomFormatter(Formatter):
    """
    Custom formatter to allow non-existing key references when formatting a
    string
    """

    _expr = r"""
        (?<!{){({{)*                # non-pair opening {
        (?:obj:)?                   # obj:
        (?P<key>\w+)                # key
        (?:\|(?P<default>[^}]*))?   # default fallback
        }(}})*(?!})                 # non-pair closing }
    """
    _matcher = re.compile(_expr, re.VERBOSE)
    _whole_matcher = re.compile(f"^{_expr}$", re.VERBOSE)

    def __init__(self, allow_empty=False):
        super().__init__()
        self.allow_empty = allow_empty

    def vformat(self, format_string, args, kwargs):
        # Special case of returning the object preserving it's type if the entire string
        # matches a single parameter.
        result = self._whole_matcher.match(format_string)
        if result is not None:
            try:
                value = kwargs[result.group("key")]
            except KeyError:
                pass
            else:
                if not isinstance(value, Undefined):
                    return value

        # handle multiple fields within string via a callback to re.sub()
        def re_replace(match):
            key = match.group("key")
            default = match.group("default")

            if default is not None:
                if key not in kwargs or isinstance(kwargs[key], Undefined):
                    return default
                else:
                    return "{%s}" % key
            return match.group(0)

        format_string = self._matcher.sub(re_replace, format_string)

        try:
            return super().vformat(format_string, args, kwargs)
        except (JenkinsJobsException, UndefinedError, ValueError) as x:
            if len(format_string) > 40:
                short_fmt = format_string[:80] + "..."
            else:
                short_fmt = format_string
            raise JenkinsJobsException(f"While formatting string {short_fmt!r}: {x}")

    def enum_required_params(self, format_string):
        def re_replace(match):
            key = match.group("key")
            return "{%s}" % key

        prepared_format_string = self._matcher.sub(re_replace, format_string)
        for literal_text, field_name, format_spec, conversion in self.parse(
            prepared_format_string
        ):
            if field_name is None:
                continue
            arg_used, rest = _string.formatter_field_name_split(field_name)
            if arg_used == "" or type(arg_used) is int:
                raise JenkinsJobsException(
                    f"Positional format arguments are not supported: {format_string!r}"
                )
            yield arg_used

    def enum_param_defaults(self, format_string):
        for match in self._matcher.finditer(format_string):
            key = match.group("key")
            default = match.group("default")
            if default is not None:
                yield (key, default)

    def get_value(self, key, args, kwargs):
        try:
            return super().get_value(key, args, kwargs)
        except KeyError:
            if self.allow_empty:
                logger.debug(
                    "Found uninitialized key %s, replaced with empty string", key
                )
                return ""
            raise JenkinsJobsException(f"Missing parameter: {key!r}")


def enum_str_format_required_params(format, pos):
    formatter = CustomFormatter()
    try:
        yield from formatter.enum_required_params(str(format))
    except JenkinsJobsException as x:
        raise x.with_pos(pos)


def enum_str_format_param_defaults(format):
    formatter = CustomFormatter()
    yield from formatter.enum_param_defaults(str(format))



import sys
from collections import namedtuple
from operator import attrgetter

import pytest

from jenkins.plugins import Plugin, PluginVersion
from jenkins_jobs.config import JJBConfig
from jenkins_jobs.registry import ModuleRegistry


Scenario = namedtuple("Scnenario", "name v1 op v2")


scenarios = [
    Scenario("s1", v1="1.0.0", op="__gt__", v2="0.8.0"),
    Scenario("s2", v1="1.0.1alpha", op="__gt__", v2="1.0.0"),
    Scenario("s3", v1="1.0", op="__eq__", v2="1.0.0"),
    Scenario("s4", v1="1.0", op="__eq__", v2="1.0"),
    Scenario("s5", v1="1.0", op="__lt__", v2="1.8.0"),
    Scenario("s6", v1="1.0.1alpha", op="__lt__", v2="1.0.1"),
    Scenario("s7", v1="1.0alpha", op="__lt__", v2="1.0.0"),
    Scenario("s8", v1="1.0-alpha", op="__lt__", v2="1.0.0"),
    Scenario("s9", v1="1.1-alpha", op="__gt__", v2="1.0"),
    Scenario("s10", v1="1.0-SNAPSHOT", op="__lt__", v2="1.0"),
    Scenario("s11", v1="1.0.preview", op="__lt__", v2="1.0"),
    Scenario("s12", v1="1.1-SNAPSHOT", op="__gt__", v2="1.0"),
    Scenario("s13", v1="1.0a-SNAPSHOT", op="__lt__", v2="1.0a"),
    Scenario(
        "s14", v1="1.4.6-SNAPSHOT (private-0986edd9-example)", op="__lt__", v2="1.4.6"
    ),
    Scenario(
        "s15", v1="1.4.6-SNAPSHOT (private-0986edd9-example)", op="__gt__", v2="1.4.5"
    ),
    Scenario("s16", v1="1.0.1-1.v1", op="__gt__", v2="1.0.1"),
    Scenario("s17", v1="1.0.1-1.v1", op="__lt__", v2="1.0.2"),
    Scenario("s18", v1="1.0.2-1.v1", op="__gt__", v2="1.0.1"),
    Scenario("s19", v1="1.0.2-1.v1", op="__gt__", v2="1.0.1-2"),
    # 'Groovy' plugin in 'inject' property.
    Scenario("s20", v1="453.vcdb_a_c5c99890", op="__ge__", v2="2.0.0"),
    # 'postbuildscript' plugin in 'postbuildscript' publisher.
    Scenario("s21", v1="3.2.0-460.va_fda_0fa_26720", op="__ge__", v2="2.0"),
    # Same, from story: 2009943.
    Scenario("s22", v1="3.1.0-375.v3db_cd92485e1", op="__ge__", v2="2.0"),
    # 'Slack Notification Plugin' in 'slack' publisher, from story: 2009819.
    Scenario("s23", v1="602.v0da_f7458945d", op="__ge__", v2="2.0"),
    # 'preSCMbuildstep' plugin in 'pre_scm_buildstep' wrapper.
    Scenario("s24", v1="44.v6ef4fd97f56e", op="__ge__", v2="0.3"),
    # 'SSH Agent Plugin' plugin in 'ssh_agent_credentials' wrapper.
    Scenario("s25", v1="295.v9ca_a_1c7cc3a_a_", op="__ge__", v2="1.5.0"),
]


@pytest.fixture(
    params=scenarios,
    ids=attrgetter("name"),
)
def scenario(request):
    return request.param


@pytest.fixture
def config():
    config = JJBConfig()
    config.validate()
    return config


@pytest.fixture
def registry(config, scenario):
    plugin_info = [
        {
            "shortName": "HerpDerpPlugin",
            "longName": "Blah Blah Blah Plugin",
        },
        {
            "shortName": "JankyPlugin1",
            "longName": "Not A Real Plugin",
            "version": scenario.v1,
        },
    ]
    return ModuleRegistry(config, [Plugin(**d) for d in plugin_info])


def test_get_plugin_version_by_short_name(scenario, registry):
    """
    Plugin version should be available by it's short name
    """
    plugin_name = "JankyPlugin1"
    version = registry.get_plugin_version(plugin_name)

    assert isinstance(version, PluginVersion)
    assert version == scenario.v1


def test_get_plugin_version_by_long_name(scenario, registry):
    """
    Plugin version should be available by it's long name
    """
    plugin_name = "Not A Real Plugin"
    version = registry.get_plugin_version(plugin_name)

    assert isinstance(version, PluginVersion)
    assert version == scenario.v1


def test_get_plugin_version_by_alternative_name(scenario, registry):
    version = registry.get_plugin_version("Non-existent name", "Not A Real Plugin")
    assert version == scenario.v1


def test_get_plugin_version_default_value(registry):
    version = registry.get_plugin_version("Non-existent name", default="1.2.3")
    assert isinstance(version, PluginVersion)
    assert version == "1.2.3"


def test_get_plugin_version_for_missing_plugin(registry):
    """
    The goal of this test case is to validate the behavior of
    ModuleRegistry.get_plugin_version when the given plugin cannot be found in
    ModuleRegistry's internal representation of the plugins_info.
    """
    plugin_name = "PluginDoesNotExist"
    version = registry.get_plugin_version(plugin_name)

    assert isinstance(version, PluginVersion)
    assert version == str(sys.maxsize)


def test_get_plugin_version_for_missing_version(registry):
    """
    The goal of this test case is to validate the behavior of
    ModuleRegistry.get_plugin_version when the given plugin shortName returns
    plugin_info dict that has no version string. In a sane world where
    plugin frameworks like Jenkins' are sane this should never happen, but
    I am including this test and the corresponding default behavior
    because, well, it's Jenkins.
    """
    plugin_name = "HerpDerpPlugin"
    version = registry.get_plugin_version(plugin_name)

    assert isinstance(version, PluginVersion)
    assert version == str(sys.maxsize)


def test_plugin_version_comparison(registry, scenario):
    """
    The goal of this test case is to validate that valid tuple versions are
    ordinally correct. That is, for each given scenario, v1.op(v2)==True
    where 'op' is the equality operator defined for the scenario.
    """
    plugin_name = "JankyPlugin1"
    v1 = registry.get_plugin_version(plugin_name)

    op = getattr(v1, scenario.op)
    test = op(scenario.v2)

    assert test, (
        f"Unexpectedly found {v1} {scenario.op} {scenario.v2} == False"
        " when comparing versions!"
    )



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import logging
import sys

from jenkins_jobs import utils
from jenkins_jobs.builder import JenkinsManager
from jenkins_jobs.errors import JenkinsJobsException
import jenkins_jobs.cli.subcommand.base as base


logger = logging.getLogger(__name__)


class DeleteAllSubCommand(base.BaseSubCommand):
    def parse_args(self, subparser):
        delete_all = subparser.add_parser(
            "delete-all",
            help="delete *ALL* jobs from Jenkins server, including "
            "those not managed by Jenkins Job Builder.",
        )

        self.parse_option_recursive_exclude(delete_all)

        delete_all.add_argument(
            "-j",
            "--jobs-only",
            action="store_true",
            dest="del_jobs",
            default=False,
            help="delete only jobs",
        )
        delete_all.add_argument(
            "-v",
            "--views-only",
            action="store_true",
            dest="del_views",
            default=False,
            help="delete only views",
        )
        delete_all.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Do not ask interactively for confirmation. Ba cautious!",
        )

    def execute(self, options, jjb_config):
        builder = JenkinsManager(jjb_config)

        reach = set()
        if options.del_jobs and options.del_views:
            raise JenkinsJobsException(
                '"--views-only" and "--jobs-only" cannot be used together.'
            )
        elif options.del_jobs and not options.del_views:
            reach.add("jobs")
        elif options.del_views and not options.del_jobs:
            reach.add("views")
        else:
            reach.update(("jobs", "views"))

        if not options.force and not utils.confirm(
            "Sure you want to delete *ALL* {} from Jenkins "
            "server?\n(including those not managed by Jenkins "
            "Job Builder)".format(" AND ".join(reach))
        ):
            sys.exit("Aborted")

        if "jobs" in reach:
            logger.info("Deleting all jobs")
            builder.delete_all_jobs()

        if "views" in reach:
            logger.info("Deleting all views")
            builder.delete_all_views()



#!/usr/bin/env python
# Copyright (C) 2015 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage Jenkins plugin module registry.

import inspect
import logging
import operator
import packaging.version
import pkg_resources
import sys
import types

from six import PY2

from jenkins.plugins import PluginVersion
from jenkins_jobs.errors import JenkinsJobsException

__all__ = ["ModuleRegistry"]

logger = logging.getLogger(__name__)

getargspec = inspect.getargspec if PY2 else inspect.getfullargspec


class ModuleRegistry(object):
    _entry_points_cache = {}
    _component_type_cache = {}

    def __init__(self, jjb_config, plugins_list=None):
        self.modules = []
        self.modules_by_component_type = {}
        self.handlers = {}
        self.jjb_config = jjb_config
        self.masked_warned = {}
        self._macros = {}

        if plugins_list is None:
            self._plugin_version = {}
        else:
            # PluginVersion by short and long plugin name.
            self._plugin_version = self._get_plugins_versions(plugins_list)

        for entrypoint in pkg_resources.iter_entry_points(group="jenkins_jobs.modules"):
            Mod = entrypoint.load()
            mod = Mod(self)
            self.modules.append(mod)
            self.modules.sort(key=operator.attrgetter("sequence"))
            if mod.component_type is not None:
                self.modules_by_component_type[mod.component_type] = entrypoint

    @staticmethod
    def _get_plugins_versions(plugins_list):
        plugin_version = {}

        for plugin_info in plugins_list:
            short_name = plugin_info.get("shortName")
            long_name = plugin_info.get("longName")

            version = plugin_info["version"]
            if version == None:  # noqa: E711; should call PluginInfo.__eq__.
                # Ensure that plugin_info always has version and it is instance of PluginVersion.
                plugin_info["version"] = str(sys.maxsize)
                version = plugin_info["version"]

            try:
                packaging.version.Version(version)
            except packaging.version.InvalidVersion:
                plugin_name = short_name or long_name
                if plugin_name:
                    logger.warning(
                        "Version %s for plugin %s does not conform to PEP440",
                        version,
                        plugin_name,
                    )
                else:
                    logger.warning("Version %s does not conform to PEP440", version)

            if short_name:
                plugin_version[short_name] = version
            if long_name:
                plugin_version[long_name] = version

        return plugin_version

    @staticmethod
    def _filter_kwargs(func, **kwargs):
        arg_spec = getargspec(func)
        for name in list(kwargs.keys()):
            if name not in arg_spec.args:
                del kwargs[name]
        return kwargs

    def get_plugin_version(self, plugin_name, alt_plugin_name=None, default=None):
        """Provide plugin version to be used from a module's impl of Base.gen_xml.

        The return value is a plugin version obtained directly from a running
        Jenkins instance.
        This allows module authors to differentiate generated XML output based
        on it.

        :arg str plugin_name: Either the shortName or longName of a plugin
          as seen in a query that looks like:
          ``http://<jenkins-hostname>/pluginManager/api/json?pretty&depth=2``
        :arg str alt_plugin_name: Alternative plugin name. Used if plugin_name
          is missing in plugin list.
        :arg str default: Default value. Used if plugin name is missing in
          plugin list.

        During a 'test' run, it is possible to override JJB's query to a live
        Jenkins instance by passing it a path to a file containing a YAML list
        of dictionaries that mimics the plugin properties you want your test
        output to reflect::

          jenkins-jobs test -p /path/to/plugins-info.yaml

        Below is example YAML that might be included in
        /path/to/plugins-info.yaml.

        .. literalinclude:: /../../tests/cmd/fixtures/plugins-info.yaml

        """
        try:
            return self._plugin_version[plugin_name]
        except KeyError:
            pass
        if alt_plugin_name:
            try:
                return self._plugin_version[alt_plugin_name]
            except KeyError:
                pass
        if default is not None:
            return PluginVersion(default)
        # Assume latest version of plugin is preferred config format.
        return PluginVersion(str(sys.maxsize))

    def registerHandler(self, category, name, method):
        cat_dict = self.handlers.get(category, {})
        if not cat_dict:
            self.handlers[category] = cat_dict
        cat_dict[name] = method

    def getHandler(self, category, name):
        return self.handlers[category][name]

    @property
    def macros(self):
        return self._macros

    def set_macros(self, macros):
        self._macros = macros

    def amend_job_dicts(self, job_data_list):
        while True:
            changed = False
            for job in job_data_list:
                for module in self.modules:
                    if module.amend_job_dict(job.data):
                        changed = True
            if not changed:
                break

    def get_component_list_type(self, entry_point):
        if entry_point in self._component_type_cache:
            return self._component_type_cache[entry_point]

        # pkg_resources.EntryPoint.load() is costly, cache it.
        component_list_type = entry_point.load().component_list_type
        logging.info("Caching type %s of %s", component_list_type, entry_point)
        self._component_type_cache[entry_point] = component_list_type

        return component_list_type

    def dispatch(
        self,
        component_type,
        xml_parent,
        component,
        template_data={},
        job_data=None,
        component_pos=None,
    ):
        """This is a method that you can call from your implementation of
        Base.gen_xml or component.  It allows modules to define a type
        of component, and benefit from extensibility via Python
        entry points and Jenkins Job Builder :ref:`Macros <macro>`.

        :arg str component_type: the name of the component
          (e.g., `builder`)
        :arg xml_parent: the parent XML element
        :arg component: component definition
        :arg dict template_data: values that should be interpolated into
          the component definition
        :arg dict job_data: full job definition

        See :py:class:`jenkins_jobs.modules.base.Base` for how to register
        components of a module.

        See the Publishers module for a simple example of how to use
        this method.
        """

        if component_type not in self.modules_by_component_type:
            raise JenkinsJobsException(
                "Unknown component type: " "'{0}'.".format(component_type)
            )

        entry_point = self.modules_by_component_type[component_type]
        component_list_type = self.get_component_list_type(entry_point)

        if isinstance(component, dict):
            # The component is a singleton dictionary of name: dict(args)
            name, component_data = next(iter(component.items()))
            if template_data:
                paramdict = {}
                paramdict.update(template_data)
                paramdict.update(job_data or {})
        else:
            # The component is a simple string name, eg "run-tests"
            name = component
            component_data = {}

        # Look for a component function defined in an entry point
        eps = self._entry_points_cache.get(component_list_type)
        if eps is None:
            eps = self._load_eps(component_list_type, component_type, entry_point, name)

        macro_dict = self.macros.get(component_type, {})
        macro = macro_dict.get(name)
        if macro:
            try:
                self._dispatch_macro(
                    component_data,
                    component_type,
                    eps,
                    job_data,
                    macro,
                    name,
                    xml_parent,
                )
            except JenkinsJobsException as x:
                if component_pos is not None:
                    raise x.with_context(
                        f"While expanding {component_type} macro call {name!r}",
                        pos=component_pos,
                    )
                else:
                    raise
        elif name in eps:
            try:
                func = eps[name]
                kwargs = self._filter_kwargs(func, job_data=job_data)
                func(self, xml_parent, component_data, **kwargs)
            except JenkinsJobsException as x:
                raise x.with_context(
                    f"In {component_type} {name!r}",
                    pos=component.pos,
                )
        else:
            raise JenkinsJobsException(
                "Unknown entry point or macro '{0}' "
                "for component type: '{1}'.".format(name, component_type)
            )

    def _dispatch_macro(
        self, component_data, component_type, eps, job_data, macro, name, xml_parent
    ):
        if name in eps and name not in self.masked_warned:
            self.masked_warned[name] = True
            logger.warning(
                "You have a macro ('%s') defined for '%s' "
                "component type that is masking an inbuilt "
                "definition" % (name, component_type)
            )
        if component_data is None:
            component_data = {}
        params = {**component_data, **(job_data or {})}
        macro.dispatch_elements(self, xml_parent, component_data, job_data, params)

    def _load_eps(self, component_list_type, component_type, entry_point, name):
        logging.debug("Caching entrypoints for %s" % component_list_type)
        module_eps = []
        # auto build entry points by inferring from base component_types
        mod = pkg_resources.EntryPoint(
            "__all__", entry_point.module_name, dist=entry_point.dist
        )
        Mod = mod.load()
        func_eps = [
            Mod.__dict__.get(a)
            for a in dir(Mod)
            if isinstance(Mod.__dict__.get(a), types.FunctionType)
        ]
        for func_ep in func_eps:
            try:
                # extract entry point based on docstring
                name_line = func_ep.__doc__.split("\n")
                if not name_line[0].startswith("yaml:"):
                    logger.debug("Ignoring '%s' as an entry point" % name_line)
                    continue
                ep_name = name_line[0].split(" ")[1]
            except (AttributeError, IndexError):
                # AttributeError by docstring not being defined as
                # a string to have split called on it.
                # IndexError raised by name_line not containing anything
                # after the 'yaml:' string.
                logger.debug(
                    "Not including func '%s' as an entry point" % func_ep.__name__
                )
                continue

            module_eps.append(
                pkg_resources.EntryPoint(
                    ep_name,
                    entry_point.module_name,
                    dist=entry_point.dist,
                    attrs=(func_ep.__name__,),
                )
            )
            logger.debug(
                "Adding auto EP '%s=%s:%s'"
                % (ep_name, entry_point.module_name, func_ep.__name__)
            )
        # load from explicitly defined entry points
        module_eps.extend(
            list(
                pkg_resources.iter_entry_points(
                    group="jenkins_jobs.{0}".format(component_list_type)
                )
            )
        )
        eps = {}
        for module_ep in module_eps:
            if module_ep.name in eps:
                raise JenkinsJobsException(
                    "Duplicate entry point found for component type: "
                    "'{0}', '{0}',"
                    "name: '{1}'".format(component_type, name)
                )

            eps[module_ep.name] = module_ep.load()
        # cache both sets of entry points
        self._entry_points_cache[component_list_type] = eps
        logger.debug("Cached entry point group %s = %s", component_list_type, eps)
        return eps



import io
from pathlib import Path

import pytest

from jenkins_jobs.cli import entry
from jenkins_jobs import builder


global_conf = "/etc/jenkins_jobs/jenkins_jobs.ini"
user_conf = Path.home() / ".config" / "jenkins_jobs" / "jenkins_jobs.ini"
local_conf = Path(__file__).parent / "jenkins_jobs.ini"


def test_use_global_config(mocker, default_config_file):
    """
    Verify that JJB uses the global config file by default
    """
    mocker.patch("jenkins_jobs.builder.JenkinsManager.get_plugins_info")

    args = ["test", "foo"]

    default_io_open = io.open

    def io_open(file, *args, **kw):
        if file == global_conf:
            default_io_open(default_config_file, "r", encoding="utf-8")
        else:
            return default_io_open(file, *args, **kw)

    def isfile(path):
        if path == global_conf:
            return True
        return False

    mocker.patch("os.path.isfile", side_effect=isfile)
    mocked_open = mocker.patch("io.open", side_effect=io_open)

    entry.JenkinsJobs(args, config_file_required=True)

    mocked_open.assert_called_with(global_conf, "r", encoding="utf-8")


def test_use_config_in_user_home(mocker, default_config_file):
    """
    Verify that JJB uses config file in user home folder
    """

    args = ["test", "foo"]

    default_io_open = io.open

    def io_open(file, *args, **kw):
        if file == str(user_conf):
            default_io_open(default_config_file, "r", encoding="utf-8")
        else:
            return default_io_open(file, *args, **kw)

    def isfile(path):
        if path == str(user_conf):
            return True
        return False

    mocker.patch("os.path.isfile", side_effect=isfile)
    mocked_open = mocker.patch("io.open", side_effect=io_open)

    entry.JenkinsJobs(args, config_file_required=True)
    mocked_open.assert_called_with(str(user_conf), "r", encoding="utf-8")


def test_non_existing_config_dir(default_config_file):
    """
    Run test mode and pass a non-existing configuration directory
    """
    args = ["--conf", default_config_file, "test", "foo"]
    jenkins_jobs = entry.JenkinsJobs(args)
    with pytest.raises(IOError):
        jenkins_jobs.execute()


def test_non_existing_config_file(default_config_file):
    """
    Run test mode and pass a non-existing configuration file
    """
    args = ["--conf", default_config_file, "test", "non-existing.yaml"]
    jenkins_jobs = entry.JenkinsJobs(args)
    with pytest.raises(IOError):
        jenkins_jobs.execute()


def test_config_options_not_replaced_by_cli_defaults(fixtures_dir):
    """
    Run test mode and check config settings from conf file retained
    when none of the global CLI options are set.
    """
    config_file = fixtures_dir / "settings_from_config.ini"
    args = ["--conf", str(config_file), "test", "dummy.yaml"]
    jenkins_jobs = entry.JenkinsJobs(args)
    jjb_config = jenkins_jobs.jjb_config
    assert jjb_config.jenkins["user"] == "jenkins_user"
    assert jjb_config.jenkins["password"] == "jenkins_password"
    assert jjb_config.builder["ignore_cache"]
    assert jjb_config.builder["flush_cache"]
    assert jjb_config.builder["update"] == "all"
    assert jjb_config.yamlparser["allow_empty_variables"]


def test_config_options_overriden_by_cli():
    """
    Run test mode and check config settings from conf file retained
    when none of the global CLI options are set.
    """
    args = [
        "--user",
        "myuser",
        "--password",
        "mypassword",
        "--ignore-cache",
        "--flush-cache",
        "--allow-empty-variables",
        "test",
        "dummy.yaml",
    ]
    jenkins_jobs = entry.JenkinsJobs(args)
    jjb_config = jenkins_jobs.jjb_config
    assert jjb_config.jenkins["user"] == "myuser"
    assert jjb_config.jenkins["password"] == "mypassword"
    assert jjb_config.builder["ignore_cache"]
    assert jjb_config.builder["flush_cache"]
    assert jjb_config.yamlparser["allow_empty_variables"]


def test_update_timeout_not_set(mocker, fixtures_dir, default_config_file):
    """Check that timeout is left unset

    Test that the Jenkins object has the timeout set on it only when
    provided via the config option.
    """
    jenkins_mock = mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager")

    path = fixtures_dir / "cmd-002.yaml"
    args = ["--conf", default_config_file, "update", str(path)]

    jenkins_mock.return_value.update_jobs.return_value = ([], 0)
    jenkins_mock.return_value.update_views.return_value = ([], 0)
    jenkins_jobs = entry.JenkinsJobs(args)
    jenkins_jobs.execute()

    # validate that the JJBConfig used to initialize builder.Jenkins
    # contains the expected timeout value.

    jjb_config = jenkins_mock.call_args[0][0]
    assert jjb_config.jenkins["timeout"] == builder._DEFAULT_TIMEOUT


def test_update_timeout_set(mocker, fixtures_dir):
    """Check that timeout is set correctly

    Test that the Jenkins object has the timeout set on it only when
    provided via the config option.
    """
    jenkins_mock = mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager")

    path = fixtures_dir / "cmd-002.yaml"
    config_file = fixtures_dir / "non-default-timeout.ini"
    args = ["--conf", str(config_file), "update", str(path)]

    jenkins_mock.return_value.update_jobs.return_value = ([], 0)
    jenkins_mock.return_value.update_views.return_value = ([], 0)
    jenkins_jobs = entry.JenkinsJobs(args)
    jenkins_jobs.execute()

    # validate that the JJBConfig used to initialize builder.Jenkins
    # contains the expected timeout value.

    jjb_config = jenkins_mock.call_args[0][0]
    assert jjb_config.jenkins["timeout"] == 0.2


def test_filter_modules_set(mocker, fixtures_dir):
    """
    Check that customs filters modules are set.

    Test that the filter_modules option is a non-empty list.
    """
    config_file = fixtures_dir / "cmd-003.conf"
    args = ["--conf", str(config_file), "test", "foo"]
    jenkins_jobs = entry.JenkinsJobs(args)

    jjb_config = jenkins_jobs.jjb_config
    assert jjb_config.yamlparser["filter_modules"] == ["my_filter", "my_other_filter"]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass

from .loc_loader import LocDict
from .position import Pos


job_contents_keys = {
    # Same as for macros.
    "parameters",
    "properties",
    "builders",
    "wrappers",
    "triggers",
    "publishers",
    "scm",
    "pipeline-scm",
    "reporters",
    # General.
    "description",
    "project-type",
    "folder",
    "node",
    "jdk",
    "actions",
    "disabled",
    "display-name",
    "block-downstream",
    "block-upstream",
    "auth-token",
    "concurrent",
    "workspace",
    "child-workspace",
    "quiet-period",
    "retry-count",
    "logrotate",
    "raw",
    # Builders.
    "prebuilders",
    "postbuilders",
    # HipChat.
    "hipchat",
    # Notificatoins.
    "notifications",
    # project Flow.
    "dsl",
    "needs-workspace",
    "dsl-file",
    # GithubOrganization.
    "prune-dead-branches",
    "days-to-keep",
    "number-to-keep",
    "periodic-folder-trigger",
    "github-org",
    "script-path",
    # Matrix.
    "execution-strategy",
    "yaml-strategy",
    "p4-strategy",
    "axes",
    # Maven.
    "maven",
    "per-module-email",
    # WorkflowMultiBranch.
    "sandbox",
    "script-id",
    "script-path",
    "prune-dead-branches",
    "days-to-keep",
    "number-to-keep",
    "abort-builds",
    "periodic-folder-trigger",
    # Pipeline.
    "dsl",
    "sandbox",
    # project Workflow.
    "dsl",
    "sandbox",
}

view_contents_keys = {
    # Common.
    "description",
    "filter-executors",
    "filter-queue",
    # All
    # <nothing>
    # List.
    "job-name",
    "job-filters",
    "width",
    "alignment",
    "columns",
    "regex",
    "recurse",
    # Sectioned.
    "sections",
    # SectionedText.
    "width",
    "alignment",
    "text",
    "style",
    # DeliveryPipeline.
    "aggregated-changes-grouping-pattern",
    "allow-abort",
    "allow-manual-triggers",
    "allow-pipeline-start",
    "allow-rebuild",
    "link-relative",
    "link-to-console-log",
    "max-number-of-visible-pipelines",
    "name",
    "no-of-columns",
    "no-of-pipelines",
    "paging-enabled",
    "show-absolute-date-time",
    "show-aggregated-changes",
    "show-aggregated-pipeline",
    "show-avatars",
    "show-changes",
    "show-description",
    "show-promotions",
    "show-static-analysis-results",
    "show-test-results",
    "show-total-build-time",
    "update-interval",
    "sorting",
    "components",
    "regexps",
    # Nested.
    "views",
    "default-view",
    "columns",
    # Pipeline.
    "first-job",
    "name",
    "no-of-displayed-builds",
    "title",
    "link-style",
    "css-Url",
    "latest-job-only",
    "manual-trigger",
    "show-parameters",
    "parameters-in-headers",
    "start-with-parameters",
    "refresh-frequency",
    "definition-header",
}


def split_contents_params(data, contents_keys):
    contents = data.copy_with(
        {key: value for key, value in data.items() if key in contents_keys}
    )
    params = data.copy_with(
        {key: value for key, value in data.items() if key not in contents_keys}
    )
    return (contents, params)


@dataclass
class Defaults:
    name: str
    pos: Pos
    params: dict
    contents: dict  # Values that go to job contents.

    @classmethod
    def add(cls, config, roots, data, pos):
        d = data.copy()
        name = d.pop_required_loc_string("name")
        contents, params = split_contents_params(
            d, job_contents_keys | view_contents_keys
        )
        defaults = cls(name, pos, params, contents)
        roots.defaults[name] = defaults

    @classmethod
    def empty(cls):
        return Defaults("empty", pos=None, params={}, contents={})

    def merged_with_global(self, global_):
        return Defaults(
            name=f"{self.name}-merged-with-global",
            pos=self.pos,
            params=LocDict.merge(global_.params, self.params),
            contents=LocDict.merge(global_.contents, self.contents),
        )



# Joint copyright:
#  - Copyright 2012,2013 Wikimedia Foundation
#  - Copyright 2012,2013 Antoine "hashar" Musso
#  - Copyright 2013 Arnaud Fabre
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from operator import attrgetter
from pathlib import Path
import sys

import pytest

from tests.enum_scenarios import scenario_list


fixtures_dir = Path(__file__).parent / "job_fixtures"


@pytest.fixture(
    params=scenario_list(fixtures_dir),
    ids=attrgetter("name"),
)
def scenario(request):
    return request.param


def test_yaml_snippet(scenario, check_job):
    old_path = sys.path
    if str(fixtures_dir) not in sys.path:
        sys.path.append(str(fixtures_dir))
    # Some tests using config with 'include_path' expect JJB root to be current directory.
    os.chdir(Path(__file__).parent / "../..")
    if scenario.name.startswith("deprecated-"):
        with pytest.warns(UserWarning) as record:
            check_job()
        assert "is deprecated" in str(record[0].message)
    else:
        check_job()
    sys.path = old_path



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The External Job Project module handles creating ExternalJob Jenkins projects.
You may specify ``externaljob`` in the ``project-type`` attribute of the
:ref:`Job` definition.

This type of job allows you to record the execution of a process run outside
Jenkins, even on a remote machine. This is designed so that you can use
Jenkins as a dashboard of your existing automation system.

Requires the Jenkins :jenkins-plugins:`External Monitor Job Type Plugin
<external-monitor-job>`.

Example:

    .. literalinclude:: /../../tests/general/fixtures/project-type005.yaml

"""

import xml.etree.ElementTree as XML

import jenkins_jobs.modules.base


class ExternalJob(jenkins_jobs.modules.base.Base):
    sequence = 0

    def root_xml(self, data):
        xml_parent = XML.Element("hudson.model.ExternalJob")
        return xml_parent



# -*- coding: utf-8 -*-
# Copyright (C) 2015 Cisco Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The folder Project module handles creating Jenkins folder projects.
You may specify ``folder`` in the ``project-type`` attribute of
the :ref:`Job` definition.

Requires the Jenkins :jenkins-plugins:`CloudBees Folders Plugin
<cloudbees-folder>`.

Job example:

    .. literalinclude::
      /../../tests/yamlparser/job_fixtures/project_folder_template001.yaml

Job template example:

    .. literalinclude::
      /../../tests/yamlparser/job_fixtures/project_folder_template002.yaml

"""

import xml.etree.ElementTree as XML
import jenkins_jobs.modules.base
from jenkins_jobs.modules.helpers import convert_mapping_to_xml


class Folder(jenkins_jobs.modules.base.Base):
    sequence = 0

    def root_xml(self, data):
        xml_parent = XML.Element(
            "com.cloudbees.hudson.plugins.folder.Folder", plugin="cloudbees-folder"
        )
        attributes = {
            "class": "com.cloudbees.hudson.plugins.folder." "icons.StockFolderIcon"
        }
        XML.SubElement(xml_parent, "icon", attrib=attributes)
        XML.SubElement(xml_parent, "views")
        attributes = {"class": "hudson.views.DefaultViewsTabBar"}
        XML.SubElement(xml_parent, "viewsTabBar", attrib=attributes)

        mappings = [("", "primaryView", "All"), ("", "healthMetrics", "")]
        convert_mapping_to_xml(xml_parent, data, mappings, True)

        return xml_parent



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Provides local yaml parsing classes and extends yaml module.

"""Custom application specific yamls tags are supported to provide
enhancements when reading yaml configuration.

Action Tags
^^^^^^^^^^^

These allow manipulation of data being stored in one layout in the source
yaml for convenience and/or clarity, to another format to be processed by
the targeted module instead of requiring all modules in JJB being capable
of supporting multiple input formats.

The tag ``!join:`` will treat the first element of the following list as
the delimiter to use, when joining the remaining elements into a string
and returning a single string to be consumed by the specified module option.

This allows users to maintain elements of data in a list structure for ease
of review/maintenance, and have the yaml parser convert it to a string for
consumption as any argument for modules. The main expected use case is to
allow for generic plugin data such as shell properties to be populated from
a list construct which the yaml parser converts to a single string, instead
of trying to support this within the module code which would require a
templating engine similar to Jinja.

Generic Example:

    .. literalinclude:: /../../tests/loader/fixtures/joinlists.yaml


Environment Inject:

    .. literalinclude:: /../../tests/yamlparser/job_fixtures/string_join.yaml


While this mechanism can also be used items where delimiters are supported by
the module, that should be considered a bug that the existing code doesn't
handle being provided a list and delimiter to perform the correct conversion
for you. Should you discover a module that takes arguments with delimiters and
the existing JJB codebase does not handle accepting lists, then this can be
used as a temporary solution in place of using very long strings:

Extended Params Example:

    .. literalinclude::
        /../../tests/parameters/fixtures/extended-choice-param-full.yaml


Inclusion Tags
^^^^^^^^^^^^^^

These allow inclusion of arbitrary files as a method of having blocks of data
managed separately to the yaml job configurations. A specific usage of this is
inlining scripts contained in separate files, although such tags may also be
used to simplify usage of macros or job templates.

The tag ``!include:`` will treat the following string as file which should be
parsed as yaml configuration data.

Example:

    .. literalinclude:: /../../tests/loader/fixtures/include001.yaml

    contents of include001.yaml.inc:

    .. literalinclude:: /../../tests/loader/fixtures/include001.yaml.inc


The tag ``!include-raw-expand:`` will treat the given string or list of strings
as filenames to be opened as one or more data blob, which should be read into
the calling yaml construct without any further parsing. Any data in a file
included through this tag, will be treated as string data.

It will expand variables inside the file. If your file contains curly braces,
you should double them. Or, you can use tag ``!include-raw-verbatim:``, which
does not substitute variables.

Examples:

    .. literalinclude::
        /../../tests/loader/fixtures/include-raw-verbatim-template.yaml

    contents of include-raw-hello-world.sh:

        .. literalinclude::
            /../../tests/loader/fixtures/include-raw-hello-world.sh

    contents of include-raw-vars.sh:

        .. literalinclude::
            /../../tests/loader/fixtures/include-raw-vars.sh

    Using a list of files:

    .. literalinclude::
        /../../tests/loader/fixtures/include-raw-verbatim-multi-template.yaml


For all the multi file includes, the files are simply appended using a newline
character.


You can also use variables in included file paths.

Example:

    .. literalinclude:: /../../tests/yamlparser/job_fixtures/lazy-load-jobs001.yaml

    with variable substitution inside included files:

    .. literalinclude:: /../../tests/yamlparser/job_fixtures/lazy-load-with-variables.yaml

    using a list of files:

    .. literalinclude::
        /../../tests/yamlparser/job_fixtures/lazy-load-jobs-multi001.yaml


The tag ``!include-jinja2:`` will treat the given string or list of strings as
filenames to be opened as Jinja2 templates, which should be rendered to a
string and included in the calling YAML construct.  (This is analogous to the
templating that will happen with ``!include-raw``.)

Examples:

    .. literalinclude:: /../../tests/yamlparser/job_fixtures/jinja01.yaml

    contents of jinja01.yaml.inc:

        .. literalinclude:: /../../tests/yamlparser/job_fixtures/jinja01.yaml.inc


The tag ``!j2:`` takes a string and treats it as a Jinja2 template.  It will be
rendered (with the variables in that context) and included in the calling YAML
construct.

Examples:

    .. literalinclude:: /../../tests/yamlparser/job_fixtures/jinja-string01.yaml

The tag ``!j2-yaml:`` is similar to the ``!j2:`` tag, just that it loads the
Jinja-rendered string as YAML and embeds it in the calling YAML construct. This
provides a very flexible and convenient way of generating pieces of YAML
structures. One of use cases is defining complex YAML structures with much
simpler configuration, without any duplication.

Examples:

    .. literalinclude:: /../../tests/yamlparser/job_fixtures/jinja-yaml01.yaml

Another use case is controlling lists dynamically, like conditionally adding
list elements based on project configuration.

Examples:

    .. literalinclude:: /../../tests/yamlparser/job_fixtures/jinja-yaml02.yaml

"""

import abc
import importlib
import logging
import traceback
import sys
from pathlib import Path

import jinja2
import jinja2.meta
import yaml

from .errors import Context, JenkinsJobsException
from .loc_loader import LocList
from .position import Pos
from .formatter import CustomFormatter, enum_str_format_required_params

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from .cached_property import cached_property

logger = logging.getLogger(__name__)


class BaseYamlObject(metaclass=abc.ABCMeta):
    @staticmethod
    def path_list_from_node(loader, node):
        if isinstance(node, yaml.ScalarNode):
            return LocList(
                [loader.construct_yaml_str(node)],
                value_pos=[loader.pos_from_node(node)],
            )
        elif isinstance(node, yaml.SequenceNode):
            return LocList(
                loader.construct_sequence(node),
                value_pos=[loader.pos_from_node(n) for n in node.value],
            )
        else:
            raise JenkinsJobsException(
                f"Expected either a sequence or scalar node, but found {node.id}",
                pos=loader.pos_from_node(node),
            )

    @classmethod
    def from_yaml(cls, loader, node):
        value = loader.construct_yaml_str(node)
        return cls(loader.jjb_config, loader, loader.pos_from_node(node), value)

    def __init__(self, jjb_config, loader, pos):
        self._search_path = jjb_config.yamlparser["include_path"].copy()
        if loader.source_dir:
            # Loaded from a file, find includes beside it too.
            self._search_path.append(loader.source_dir)
        self._filter_modules = jjb_config.yamlparser["filter_modules"].copy()
        self._loader = loader
        self._pos = pos
        allow_empty = jjb_config.yamlparser["allow_empty_variables"]
        self._formatter = CustomFormatter(allow_empty)

    @abc.abstractmethod
    def expand(self, expander, params):
        """Expand object and substitute template parameters"""
        pass

    def _find_file(self, rel_path, pos):
        search_path = self._search_path
        if "." not in search_path:
            search_path.append(".")
        dir_list = [Path(d).expanduser() for d in self._search_path]
        for dir in dir_list:
            candidate = dir.joinpath(rel_path)
            if candidate.is_file():
                logger.debug("Including file %r from path %r", str(rel_path), str(dir))
                return candidate
        dir_list_str = ",".join(str(d) for d in dir_list)
        raise JenkinsJobsException(
            f"File {rel_path} does not exist in any of include directories: {dir_list_str}",
            pos=pos,
        )

    def _expand_path_list(self, path_list, *args):
        for idx, path in enumerate(path_list):
            yield self._expand_path(path, path_list.value_pos[idx], *args)


class J2BaseYamlObject(BaseYamlObject):
    def __init__(self, jjb_config, loader, pos):
        super().__init__(jjb_config, loader, pos)
        self._filters = {}
        for module_name in self._filter_modules:
            module = importlib.import_module(module_name)
            self._filters.update(module.FILTERS)
        self._jinja2_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self._search_path),
            undefined=jinja2.StrictUndefined,
        )
        self._jinja2_env.filters.update(self._filters)

    def _render_template(self, pos, template_text, template, params):
        try:
            return template.render(params)
        except jinja2.UndefinedError as x:
            # Jinja2 adds fake traceback entry with template line number.
            tb = traceback.extract_tb(x.__traceback__)
            line_ofs = tb[-1].lineno - 1  # traceback lineno starts with 1.
            lines = template_text.splitlines()
            start_ofs = pos.body.index(lines[0])
            # Examples for pre_pad: '!j2: \n<indent spaces>', '!j2: '.
            pre_pad = pos.body[:start_ofs]
            # Shift position to reflect template position inside yaml file:
            if "\n" in pre_pad:
                pos = pos.with_offset(line_ofs=1)
                column_ofs = 0
            else:
                column_ofs = start_ofs
            # Move position to error inside template:
            pos = pos.with_offset(line_ofs, column_ofs)
            pos = pos.with_contents_start()
            if len(template_text) > 40:
                text = template_text[:40] + "..."
            else:
                text = template_text
            context = Context(f"While formatting jinja2 template {text!r}", self._pos)
            raise JenkinsJobsException(str(x), pos=pos, ctx=[context])


class J2Template(J2BaseYamlObject):
    def __init__(self, jjb_config, loader, pos, template_text):
        super().__init__(jjb_config, loader, pos)
        self._template_text = template_text
        self._template = self._jinja2_env.from_string(template_text)

    def _params_from_referenced_templates(self, template_text):
        """
        Find recursively undeclared jinja2 variables from any
        (nested) included template(s)
        """
        required_params = set()
        ast = self._jinja2_env.parse(template_text)
        for rt in jinja2.meta.find_referenced_templates(ast):
            # recursive call to find params also from nested includes
            template_text = Path(self._find_file(rt, 0)).read_text()
            required_params.update(
                self._params_from_referenced_templates(template_text)
            )
        required_params.update(jinja2.meta.find_undeclared_variables(ast))
        return required_params

    @cached_property
    def required_params(self):
        return self._params_from_referenced_templates(self._template_text)

    def _render(self, params):
        return self._render_template(
            self._pos, self._template_text, self._template, params
        )


class J2String(J2Template):
    yaml_tag = "!j2:"

    def expand(self, expander, params):
        return self._render(params)


class J2Yaml(J2Template):
    yaml_tag = "!j2-yaml:"

    def expand(self, expander, params):
        text = self._render(params)
        data = self._loader.load(
            text, source_path="<expanded j2-yaml>", source_dir=self._loader.source_dir
        )
        try:
            return expander.expand(data, params)
        except JenkinsJobsException as x:
            raise x.with_context("In expanded !j2-yaml:", self._pos)


class IncludeJinja2(J2BaseYamlObject):
    yaml_tag = "!include-jinja2:"

    @classmethod
    def from_yaml(cls, loader, node):
        path_list = cls.path_list_from_node(loader, node)
        return cls(loader.jjb_config, loader, loader.pos_from_node(node), path_list)

    def __init__(self, jjb_config, loader, pos, path_list):
        super().__init__(jjb_config, loader, pos)
        self._path_list = path_list

    @property
    def required_params(self):
        return []

    def expand(self, expander, params):
        return "\n".join(self._expand_path_list(self._path_list, expander, params))

    def _expand_path(self, path_template, pos, expander, params):
        rel_path = self._formatter.format(path_template, **params)
        full_path = self._find_file(rel_path, pos)
        template_text = full_path.read_text()
        template = self._jinja2_env.from_string(template_text)
        pos = Pos.from_file(full_path, template_text)
        try:
            return self._render_template(pos, template_text, template, params)
        except JenkinsJobsException as x:
            raise x.with_context(f"In included file {str(full_path)!r}", pos=self._pos)


class IncludeBaseObject(BaseYamlObject):
    @classmethod
    def from_yaml(cls, loader, node):
        path_list = cls.path_list_from_node(loader, node)
        return cls(loader.jjb_config, loader, loader.pos_from_node(node), path_list)

    def __init__(self, jjb_config, loader, pos, path_list):
        super().__init__(jjb_config, loader, pos)
        self._path_list = path_list

    @property
    def required_params(self):
        for idx, path in enumerate(self._path_list):
            yield from enum_str_format_required_params(
                path, pos=self._path_list.value_pos[idx]
            )


class YamlInclude(IncludeBaseObject):
    yaml_tag = "!include:"

    def expand(self, expander, params):
        yaml_list = list(self._expand_path_list(self._path_list, expander, params))
        if len(yaml_list) == 1:
            return yaml_list[0]
        else:
            return "\n".join(yaml_list)

    def _expand_path(self, path_template, pos, expander, params):
        rel_path = self._formatter.format(path_template, **params)
        full_path = self._find_file(rel_path, pos)
        data = self._loader.load_path(full_path)
        try:
            return expander.expand(data, params)
        except JenkinsJobsException as x:
            raise x.with_context(f"In included file {str(full_path)!r}", pos=self._pos)


class IncludeRawBase(IncludeBaseObject):
    def expand(self, expander, params):
        return "\n".join(self._expand_path_list(self._path_list, params))


class IncludeRawExpand(IncludeRawBase):
    yaml_tag = "!include-raw-expand:"

    def _expand_path(self, rel_path_template, pos, params):
        rel_path = self._formatter.format(rel_path_template, **params)
        full_path = self._find_file(rel_path, pos)
        template = full_path.read_text()
        try:
            return self._formatter.format(template, **params)
        except JenkinsJobsException as x:
            raise x.with_context(f"In included file {str(full_path)!r}", pos=self._pos)


class IncludeRawVerbatim(IncludeRawBase):
    yaml_tag = "!include-raw-verbatim:"

    def _expand_path(self, rel_path_template, pos, params):
        rel_path = self._formatter.format(rel_path_template, **params)
        full_path = self._find_file(rel_path, pos)
        return full_path.read_text()


class YamlListJoin:
    yaml_tag = "!join:"

    @classmethod
    def from_yaml(cls, loader, node):
        value = loader.construct_sequence(node, deep=True)
        if len(value) != 2:
            raise JenkinsJobsException(
                "Join value should contain 2 elements: delimiter and string list,"
                f" but contains {len(value)} elements: {value!r}",
                pos=loader.pos_from_node(node),
            )
        delimiter, seq = value
        return delimiter.join(seq)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass

from .errors import JenkinsJobsException
from .loc_loader import LocDict
from .expander import Expander
from .root_base import RootBase, NonTemplateRootMixin, TemplateRootMixin, Group
from .defaults import split_contents_params, view_contents_keys


@dataclass
class ViewBase(RootBase):
    view_type: str

    @classmethod
    def from_dict(cls, config, roots, data, pos):
        keep_descriptions = config.yamlparser["keep_descriptions"]
        d = data.copy()
        name = d.pop_required_loc_string("name")
        id = d.pop_loc_string("id", None)
        description = d.pop_loc_string("description", None)
        defaults = d.pop_loc_string("defaults", "global")
        view_type = d.pop_loc_string("view-type", "list")
        contents, params = split_contents_params(d, view_contents_keys)
        return cls(
            _defaults=roots.defaults,
            _expander=Expander(config),
            _keep_descriptions=keep_descriptions,
            _id=id,
            name=name,
            pos=pos,
            description=description,
            defaults_name=defaults,
            params=params,
            _contents=contents,
            view_type=view_type,
        )

    @property
    def contents(self):
        return LocDict.merge(
            {
                "name": self.name,
                "view-type": self.view_type,
            },
            super().contents,
        )


class View(ViewBase, NonTemplateRootMixin):
    @classmethod
    def add(cls, config, roots, data, pos):
        view = cls.from_dict(config, roots, data, pos)
        roots.assign(roots.views, view.id, view, "view")

    def __str__(self):
        return f"view {self.name!r}"


class ViewTemplate(ViewBase, TemplateRootMixin):
    @classmethod
    def add(cls, config, roots, data, pos):
        template = cls.from_dict(config, roots, data, pos)
        roots.assign(roots.view_templates, template.id, template, "view template")

    def __str__(self):
        return f"view template {self.name!r}"


@dataclass
class ViewGroup(Group):
    _views: dict
    _view_templates: dict

    @classmethod
    def add(cls, config, roots, data, pos):
        d = data.copy()
        name = d.pop_required_loc_string("name")
        try:
            view_specs = cls._specs_from_list(d.pop("views", None))
        except JenkinsJobsException as x:
            raise x.with_context(f"In view {name!r}", pos=pos)
        group = cls(
            name,
            pos,
            view_specs,
            d,
            roots.views,
            roots.view_templates,
        )
        roots.assign(roots.view_groups, group.name, group, "view group")

    def __str__(self):
        return f"view group {self.name!r}"

    @property
    def _root_dicts(self):
        return [self._views, self._view_templates]



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Base class for a jenkins_jobs module

from jenkins_jobs.loc_loader import LocList


class Base(object):
    """
    A base class for a Jenkins Job Builder Module.

    The module is initialized before any YAML is parsed.

    :arg ModuleRegistry registry: the global module registry.
    """

    #: The sequence number for the module.  Modules are invoked in the
    #: order of their sequence number in order to produce consistently
    #: ordered XML output.
    sequence = 10

    #: The component type for components of this module.  This will be
    #: used to look for macros (they are defined singularly, and should
    #: not be plural).
    #: Set both component_type and component_list_type to None if module
    #: doesn't have components.
    component_type = None

    #: The component list type will be used to look up possible
    #: implementations of the component type via entry points (entry
    #: points provide a list of components, so it should be plural).
    #: Set both component_type and component_list_type to None if module
    #: doesn't have components.
    component_list_type = None

    def __init__(self, registry):
        self.registry = registry

    def amend_job_dict(self, job):
        """This method is called before any XML is generated.  By
        overriding this method, a module may arbitrarily modify a job data
        structure which will probably be the JJB Job intermediate data dict
        representation. If it has changed the data structure at all, it must
        return ``True``, otherwise, it must return ``False``.

        :arg dict job: the intermediate representation of job data
            loaded from JJB Yaml files after variables interpolation and other
            yaml expansions.

        :rtype: bool
        """

        return False

    def gen_xml(self, xml_parent, data):
        """Update the XML element tree based on YAML data.  Override
        this method to add elements to the XML output.  Create new
        Element objects and add them to the xml_parent.  The YAML data
        structure must not be modified.

        :arg class:`xml.etree.ElementTree` xml_parent: the parent XML element
        :arg dict data: the YAML data structure
        """

        pass

    def dispatch_component_list(
        self, component_type, component_list, xml_parent, job_data=None
    ):
        if not component_list:
            return
        for idx, component in enumerate(component_list):
            if isinstance(component_list, LocList):
                pos = component_list.value_pos[idx]
            else:
                pos = None
            self.registry.dispatch(
                component_type,
                xml_parent,
                component,
                job_data=job_data,
                component_pos=pos,
            )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass

from .errors import JenkinsJobsException
from .position import Pos
from .root_base import GroupBase


@dataclass
class Project(GroupBase):
    _jobs: dict
    _job_templates: dict
    _job_groups: dict
    _views: dict
    _view_templates: dict
    _view_groups: dict
    name: str
    pos: Pos
    defaults_name: str
    job_specs: list  # list[Spec]
    view_specs: list  # list[Spec]
    params: dict

    @classmethod
    def add(cls, config, roots, data, pos):
        d = data.copy()
        name = d.pop_required_loc_string("name")
        defaults = d.pop_loc_string("defaults", None)
        try:
            job_specs = cls._specs_from_list(d.pop("jobs", None))
            view_specs = cls._specs_from_list(d.pop("views", None))
        except JenkinsJobsException as x:
            raise x.with_context(f"In project {name!r}", pos=pos)
        project = cls(
            roots.jobs,
            roots.job_templates,
            roots.job_groups,
            roots.views,
            roots.view_templates,
            roots.view_groups,
            name,
            pos,
            defaults,
            job_specs,
            view_specs,
            params=d,
        )
        roots.assign(roots.projects, project.name, project, "project")

    def __str__(self):
        return f"project {self.name!r}"

    @property
    def _my_params(self):
        return {"name": self.name}

    def generate_jobs(self):
        root_dicts = [self._jobs, self._job_templates, self._job_groups]
        return self._generate_items(
            root_dicts, self.job_specs, self.defaults_name, params={}
        )

    def generate_views(self):
        root_dicts = [self._views, self._view_templates, self._view_groups]
        return self._generate_items(
            root_dicts, self.view_specs, self.defaults_name, params={}
        )



# Joint copyright:
#  - Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# The goal of these tests is to check that given a particular set of flags to
# Jenkins Job Builder's command line tools it will result in a particular set
# of actions by the JJB library, usually through interaction with the
# python-jenkins library.

from unittest import mock


def test_delete_single_job(mocker, default_config_file, execute_jenkins_jobs):
    """
    Test handling the deletion of a single Jenkins job.
    """

    mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_jobs")
    mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_views")

    args = ["--conf", default_config_file, "delete", "test_job"]
    execute_jenkins_jobs(args)


def test_delete_multiple_jobs(mocker, default_config_file, execute_jenkins_jobs):
    """
    Test handling the deletion of multiple Jenkins jobs.
    """

    mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_jobs")
    mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_views")

    args = ["--conf", default_config_file, "delete", "test_job1", "test_job2"]
    execute_jenkins_jobs(args)


def test_delete_using_glob_params(
    mocker, fixtures_dir, default_config_file, execute_jenkins_jobs
):
    """
    Test handling the deletion of multiple Jenkins jobs using the glob
    parameters feature.
    """

    delete_job_mock = mocker.patch("jenkins_jobs.builder.JenkinsManager.delete_job")

    args = [
        "--conf",
        default_config_file,
        "delete",
        "--path",
        str(fixtures_dir / "cmd-002.yaml"),
        "*bar*",
    ]
    execute_jenkins_jobs(args)
    calls = [mock.call("bar001"), mock.call("bar002")]
    delete_job_mock.assert_has_calls(calls, any_order=True)
    assert delete_job_mock.call_count == len(
        calls
    ), "Jenkins.delete_job() was called '%s' times when " "expected '%s'" % (
        delete_job_mock.call_count,
        len(calls),
    )



# Joint copyright:
#  - Copyright 2012,2013 Wikimedia Foundation
#  - Copyright 2012,2013 Antoine "hashar" Musso
#  - Copyright 2013 Arnaud Fabre
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from operator import attrgetter
from pathlib import Path

import pytest

from jenkins_jobs.errors import JenkinsJobsException
from tests.enum_scenarios import scenario_list

fixtures_dir = Path(__file__).parent / "error_fixtures"


@pytest.fixture(
    params=scenario_list(fixtures_dir),
    ids=attrgetter("name"),
)
def scenario(request):
    return request.param


# Override to avoid scenarios usage.
@pytest.fixture
def config_path():
    return os.devnull


# Override to avoid scenarios usage.
@pytest.fixture
def plugins_info():
    return None


def test_error(check_parser, scenario, expected_error):
    with pytest.raises(JenkinsJobsException) as excinfo:
        check_parser(scenario.in_path)
    error = "\n".join(excinfo.value.lines)
    print()
    print(error)
    canonical_error = error.replace(str(fixtures_dir) + "/", "").replace(
        str(fixtures_dir), "fixtures-dir"
    )
    assert canonical_error == expected_error



from operator import attrgetter
from pathlib import Path

import pytest

from jenkins_jobs.errors import JenkinsJobsException
from tests.enum_scenarios import scenario_list


fixtures_dir = Path(__file__).parent / "view_fixtures"


@pytest.fixture(
    params=scenario_list(fixtures_dir),
    ids=attrgetter("name"),
)
def scenario(request):
    return request.param


def test_yaml_snippet(scenario, expected_error, check_view):
    if scenario.error_path.exists():
        with pytest.raises(JenkinsJobsException) as excinfo:
            check_view()
        error = "\n".join(excinfo.value.lines)
        print()
        print(error)
        assert error.replace(str(fixtures_dir) + "/", "") == expected_error
    else:
        check_view()



# -*- coding: utf-8 -*-
# Copyright (C) 2016 Mirantis, Inc.
#
# Based on jenkins_jobs/modules/project_workflow.py by
# Copyright (C) 2015 David Caro <david@dcaro.es>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The Pipeline Project module handles creating Jenkins Pipeline projects
(formerly known as the Workflow projects).
You may specify ``pipeline`` in the ``project-type`` attribute of
the :ref:`Job` definition.

Requires the Jenkins :jenkins-plugins:`Pipeline Plugin <workflow-aggregator>`.

In order to write an inline script within a job-template you have to escape the
curly braces by doubling them in the DSL: { -> {{ , otherwise it will be
interpreted by the python str.format() command.

:Job Parameters:
    * **sandbox** (`bool`): If the script should run in a sandbox (default
      false)
    * **dsl** (`str`): The DSL content or,
    * **pipeline-scm** (`str`): in case "Pipeline as code" feature is used.
      Then you should specify:

        * **scm**: single ``scm`` component (or a reference) describing the
          source code repository
        * **script-path**: path to the Groovy file containing the job's steps
          (optional, default: ``Jenkinsfile``)
        * **lightweight-checkout** (`bool`): If selected, try to obtain the
          Pipeline script contents directly from the SCM without performing a
          full checkout. (optional, default: ``false``)

Note that ``dsl`` and ``pipeline-scm`` parameters are mutually exclusive.

Inline DSL job example:

    .. literalinclude::
      /../../tests/yamlparser/job_fixtures/project_pipeline_template001.yaml

Inline DSL job template example:

    .. literalinclude::
      /../../tests/yamlparser/job_fixtures/project_pipeline_template002.yaml

"Pipeline as code" example:

    .. literalinclude::
      /../../tests/yamlparser/job_fixtures/project_pipeline_template004.yaml

"Pipeline as code" example using templates:

    .. literalinclude::
      /../../tests/yamlparser/job_fixtures/project_pipeline_template005.yaml

"Pipeline as nested stage" example :

    .. literalinclude::
      /../../tests/yamlparser/job_fixtures/project_pipeline_template006.yaml

.. _Pipeline as code: https://jenkins.io/solutions/pipeline/

"""
import xml.etree.ElementTree as XML

from jenkins_jobs.errors import JenkinsJobsException
import jenkins_jobs.modules.base


class Pipeline(jenkins_jobs.modules.base.Base):
    sequence = 0
    error_msg = "You cannot declare both 'dsl' and 'pipeline-scm' on a " "pipeline job"

    def root_xml(self, data):
        xml_parent = XML.Element("flow-definition", {"plugin": "workflow-job"})
        if "dsl" in data and "pipeline-scm" in data:
            raise JenkinsJobsException(self.error_msg)
        if "dsl" in data:
            xml_definition = XML.SubElement(
                xml_parent,
                "definition",
                {
                    "class": "org.jenkinsci.plugins." "workflow.cps.CpsFlowDefinition",
                    "plugin": "workflow-cps",
                },
            )
            XML.SubElement(xml_definition, "script").text = data["dsl"]
        elif "pipeline-scm" in data:
            xml_definition = XML.SubElement(
                xml_parent,
                "definition",
                {
                    "class": "org.jenkinsci.plugins.workflow.cps."
                    "CpsScmFlowDefinition",
                    "plugin": "workflow-cps",
                },
            )
        else:
            raise JenkinsJobsException(
                "Either 'dsl' or 'pipeline-scm' " "is required for pipeline job"
            )

        needs_workspace = data.get("sandbox", False)
        XML.SubElement(xml_definition, "sandbox").text = str(needs_workspace).lower()

        return xml_parent



#  - Copyright 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# The goal of these tests is to check that given a particular set of flags to
# Jenkins Job Builder's command line tools it will result in a particular set
# of actions by the JJB library, usually through interaction with the
# python-jenkins library.

import pytest


def test_delete_all_accept(mocker, default_config_file, execute_jenkins_jobs):
    """
    Test handling the deletion of a single Jenkins job.
    """

    mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_all_jobs")
    mocker.patch("jenkins_jobs.builder.JenkinsManager.get_views", return_value=[None])
    mocker.patch("jenkins_jobs.utils.input", return_value="y")

    args = ["--conf", default_config_file, "delete-all"]
    execute_jenkins_jobs(args)


def test_delete_all_abort(mocker, default_config_file, execute_jenkins_jobs):
    """
    Test handling the deletion of a single Jenkins job.
    """

    mocker.patch("jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_all_jobs")
    mocker.patch("jenkins_jobs.utils.input", return_value="n")

    args = ["--conf", default_config_file, "delete-all"]
    with pytest.raises(SystemExit):
        execute_jenkins_jobs(args)


def test_delete_all_forced(mocker, default_config_file, execute_jenkins_jobs):
    """
    Test handling the deletion of a job and a view with --force flag.
    """

    delete_jobs = mocker.patch(
        "jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_all_jobs"
    )
    delete_views = mocker.patch(
        "jenkins_jobs.cli.subcommand.base.JenkinsManager.delete_all_views"
    )
    get_jobs = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_all_jobs")
    get_jobs.return_value = [{"name": name} for name in ["job-1"]]
    get_views = mocker.patch("jenkins_jobs.builder.jenkins.Jenkins.get_views")
    get_views.return_value = [{"name": name} for name in ["view-1"]]
    input = mocker.patch("jenkins_jobs.utils.input", return_value="n")

    args = ["--conf", default_config_file, "delete-all", "--force"]
    execute_jenkins_jobs(args)

    input.assert_not_called()

    assert delete_jobs.call_count == 1
    assert delete_views.call_count == 1



#!/usr/bin/env python
#
# Joint copyright:
#  - Copyright 2012,2013 Wikimedia Foundation
#  - Copyright 2012,2013 Antoine "hashar" Musso
#  - Copyright 2013 Arnaud Fabre
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple


Scenario = namedtuple(
    "Scenario", "name in_path out_paths error_path config_path plugins_info_path"
)


def scenario_list(fixtures_dir, in_ext=".yaml", out_ext=".xml"):
    compound_dirs = set()
    if out_ext == ".xml":
        for path in fixtures_dir.rglob("expected-output.xml"):
            dir = path.parent
            compound_dirs.add(dir)
            yield Scenario(
                name=dir.stem,
                in_path=dir,
                out_paths=[path],
                error_path=dir / "expected.error",
                # When config file is missing it will still be passed and not None,
                # so JJBConfig will prefer it over system and user configs.
                config_path=dir / "test.conf",
                plugins_info_path=dir / "test.plugins_info.yaml",
            )
    for path in fixtures_dir.rglob(f"*{in_ext}"):
        if path.name.endswith("plugins_info.yaml"):
            continue
        if any(d in path.parents for d in compound_dirs):
            continue
        out_path = path.with_suffix(out_ext)
        out_path_list = list(fixtures_dir.rglob(out_path.name))
        yield Scenario(
            name=path.stem,
            in_path=path,
            out_paths=out_path_list,
            error_path=path.with_suffix(".error"),
            # When config file is missing it will still be passed and not None,
            # so JJBConfig will prefer it over system and user configs.
            config_path=path.with_suffix(".conf"),
            plugins_info_path=path.with_suffix(".plugins_info.yaml"),
        )



# Copyright 2015 Thanh Ha
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from functools import wraps
import logging

import xml.etree.ElementTree as XML

from jenkins_jobs.errors import InvalidAttributeError
from jenkins_jobs.errors import JenkinsJobsException
from jenkins_jobs.errors import MissingAttributeError
from jenkins_jobs.modules import hudson_model


def build_trends_publisher(plugin_name, xml_element, data):
    """Helper to create various trend publishers."""

    def append_thresholds(element, data, only_totals):
        """Appends the status thresholds."""

        for status in ["unstable", "failed"]:
            status_data = data.get(status, {})

            limits = [
                ("total-all", "TotalAll"),
                ("total-high", "TotalHigh"),
                ("total-normal", "TotalNormal"),
                ("total-low", "TotalLow"),
            ]

            if only_totals is False:
                limits.extend(
                    [
                        ("new-all", "NewAll"),
                        ("new-high", "NewHigh"),
                        ("new-normal", "NewNormal"),
                        ("new-low", "NewLow"),
                    ]
                )

            for key, tag_suffix in limits:
                tag_name = status + tag_suffix
                XML.SubElement(element, tag_name).text = str(status_data.get(key, ""))

    # Tuples containing: setting name, tag name, default value
    settings = [
        ("healthy", "healthy", ""),
        ("unhealthy", "unHealthy", ""),
        ("health-threshold", "thresholdLimit", "low"),
        ("plugin-name", "pluginName", plugin_name),
        ("default-encoding", "defaultEncoding", ""),
        ("can-run-on-failed", "canRunOnFailed", False),
        ("use-stable-build-as-reference", "useStableBuildAsReference", False),
        ("use-previous-build-as-reference", "usePreviousBuildAsReference", False),
        ("use-delta-values", "useDeltaValues", False),
        ("thresholds", "thresholds", {}),
        ("should-detect-modules", "shouldDetectModules", False),
        ("dont-compute-new", "dontComputeNew", True),
        ("do-not-resolve-relative-paths", "doNotResolveRelativePaths", False),
        ("pattern", "pattern", ""),
    ]

    thresholds = ["low", "normal", "high"]

    for key, tag_name, default in settings:
        xml_config = XML.SubElement(xml_element, tag_name)
        config_value = data.get(key, default)

        if key == "thresholds":
            append_thresholds(
                xml_config, config_value, data.get("dont-compute-new", True)
            )
        elif key == "health-threshold" and config_value not in thresholds:
            raise JenkinsJobsException(
                "health-threshold must be one of %s" % ", ".join(thresholds)
            )
        else:
            if isinstance(default, bool):
                xml_config.text = str(config_value).lower()
            else:
                xml_config.text = str(config_value)


def config_file_provider_builder(xml_parent, data):
    """Builder / Wrapper helper"""
    xml_files = XML.SubElement(xml_parent, "managedFiles")

    files = data.get("files", [])
    for file in files:
        xml_file = XML.SubElement(
            xml_files, "org.jenkinsci.plugins." "configfiles.buildwrapper.ManagedFile"
        )
        mapping = [
            ("file-id", "fileId", None),
            ("target", "targetLocation", ""),
            ("variable", "variable", ""),
            ("replace-tokens", "replaceTokens", False),
        ]
        convert_mapping_to_xml(xml_file, file, mapping, fail_required=True)


def config_file_provider_settings(xml_parent, data):
    SETTINGS_TYPES = ["file", "cfp"]
    settings = {
        "default-settings": "jenkins.mvn.DefaultSettingsProvider",
        "settings": "jenkins.mvn.FilePathSettingsProvider",
        "config-file-provider-settings": "org.jenkinsci.plugins.configfiles.maven.job.MvnSettingsProvider",
        "default-global-settings": "jenkins.mvn.DefaultGlobalSettingsProvider",
        "global-settings": "jenkins.mvn.FilePathGlobalSettingsProvider",
        "config-file-provider-global-settings": "org.jenkinsci.plugins.configfiles.maven.job."
        "MvnGlobalSettingsProvider",
    }

    if "settings" in data:
        # Support for Config File Provider
        settings_file = str(data["settings"])
        settings_type = data.get("settings-type", "file")

        # For cfp versions <2.10.0 we are able to detect cfp via the config
        # settings name.
        text = "org.jenkinsci.plugins.configfiles.maven.MavenSettingsConfig"
        if settings_file.startswith(text):
            settings_type = "cfp"

        if settings_type == "file":
            lsettings = XML.SubElement(
                xml_parent, "settings", {"class": settings["settings"]}
            )
            XML.SubElement(lsettings, "path").text = settings_file
        elif settings_type == "cfp":
            lsettings = XML.SubElement(
                xml_parent,
                "settings",
                {"class": settings["config-file-provider-settings"]},
            )
            XML.SubElement(lsettings, "settingsConfigId").text = settings_file
        else:
            raise InvalidAttributeError("settings-type", settings_type, SETTINGS_TYPES)
    else:
        XML.SubElement(xml_parent, "settings", {"class": settings["default-settings"]})

    if "global-settings" in data:
        # Support for Config File Provider
        global_settings_file = str(data["global-settings"])
        global_settings_type = data.get("global-settings-type", "file")

        # For cfp versions <2.10.0 we are able to detect cfp via the config
        # settings name.
        text = "org.jenkinsci.plugins.configfiles.maven." "GlobalMavenSettingsConfig"
        if global_settings_file.startswith(text):
            global_settings_type = "cfp"

        if global_settings_type == "file":
            gsettings = XML.SubElement(
                xml_parent, "globalSettings", {"class": settings["global-settings"]}
            )
            XML.SubElement(gsettings, "path").text = global_settings_file
        elif global_settings_type == "cfp":
            gsettings = XML.SubElement(
                xml_parent,
                "globalSettings",
                {"class": settings["config-file-provider-global-settings"]},
            )
            XML.SubElement(gsettings, "settingsConfigId").text = global_settings_file
        else:
            raise InvalidAttributeError(
                "settings-type", global_settings_type, SETTINGS_TYPES
            )
    else:
        XML.SubElement(
            xml_parent, "globalSettings", {"class": settings["default-global-settings"]}
        )


def copyartifact_build_selector(xml_parent, data, select_tag="selector"):

    select = data.get("which-build", "last-successful")
    selectdict = {
        "last-successful": "StatusBuildSelector",
        "last-completed": "LastCompletedBuildSelector",
        "specific-build": "SpecificBuildSelector",
        "last-saved": "SavedBuildSelector",
        "upstream-build": "TriggeredBuildSelector",
        "permalink": "PermalinkBuildSelector",
        "workspace-latest": "WorkspaceSelector",
        "build-param": "ParameterizedBuildSelector",
        "downstream-build": "DownstreamBuildSelector",
        "multijob-build": "MultiJobBuildSelector",
    }
    if select not in selectdict:
        raise InvalidAttributeError("which-build", select, selectdict.keys())
    permalink = data.get("permalink", "last")
    permalinkdict = {
        "last": "lastBuild",
        "last-stable": "lastStableBuild",
        "last-successful": "lastSuccessfulBuild",
        "last-failed": "lastFailedBuild",
        "last-unstable": "lastUnstableBuild",
        "last-unsuccessful": "lastUnsuccessfulBuild",
    }
    if permalink not in permalinkdict:
        raise InvalidAttributeError("permalink", permalink, permalinkdict.keys())
    if select == "multijob-build":
        selector = XML.SubElement(
            xml_parent,
            select_tag,
            {"class": "com.tikal.jenkins.plugins.multijob." + selectdict[select]},
        )
    else:
        selector = XML.SubElement(
            xml_parent,
            select_tag,
            {"class": "hudson.plugins.copyartifact." + selectdict[select]},
        )
    mapping = []
    if select == "specific-build":
        mapping.append(("build-number", "buildNumber", ""))
    if select == "last-successful":
        mapping.append(("stable", "stable", False))
    if select == "upstream-build":
        mapping.append(
            ("fallback-to-last-successful", "fallbackToLastSuccessful", False)
        )
    if select == "permalink":
        mapping.append(("", "id", permalinkdict[permalink]))
    if select == "build-param":
        mapping.append(("param", "parameterName", ""))
    if select == "downstream-build":
        mapping.append(("upstream-project-name", "upstreamProjectName", ""))
        mapping.append(("upstream-build-number", "upstreamBuildNumber", ""))
    convert_mapping_to_xml(selector, data, mapping, fail_required=False)


def findbugs_settings(xml_parent, data):
    # General Options
    mapping = [
        ("rank-priority", "isRankActivated", False),
        ("include-files", "includePattern", ""),
        ("exclude-files", "excludePattern", ""),
    ]
    convert_mapping_to_xml(xml_parent, data, mapping, fail_required=True)


def get_value_from_yaml_or_config_file(key, section, data, jjb_config):
    return jjb_config.get_plugin_config(section, key, data.get(key, ""))


def cloudformation_region_dict():
    region_dict = {
        "us-east-1": "US_East_Northern_Virginia",
        "us-west-1": "US_WEST_Northern_California",
        "us-west-2": "US_WEST_Oregon",
        "eu-central-1": "EU_Frankfurt",
        "eu-west-1": "EU_Ireland",
        "ap-southeast-1": "Asia_Pacific_Singapore",
        "ap-southeast-2": "Asia_Pacific_Sydney",
        "ap-northeast-1": "Asia_Pacific_Tokyo",
        "sa-east-1": "South_America_Sao_Paulo",
    }
    return region_dict


def cloudformation_init(xml_parent, data, xml_tag):
    cloudformation = XML.SubElement(
        xml_parent,
        "com.syncapse.jenkinsci." "plugins.awscloudformationwrapper." + xml_tag,
    )
    return XML.SubElement(cloudformation, "stacks")


def cloudformation_stack(xml_parent, stack, xml_tag, stacks, region_dict):
    if "name" not in stack or stack["name"] == "":
        raise MissingAttributeError("name")
    step = XML.SubElement(
        stacks, "com.syncapse.jenkinsci.plugins." "awscloudformationwrapper." + xml_tag
    )

    if xml_tag == "SimpleStackBean":
        mapping = [("prefix", "isPrefixSelected", False)]
    else:
        parameters_value = ",".join(stack.get("parameters", []))
        mapping = [
            ("description", "description", ""),
            ("", "parameters", parameters_value),
            ("timeout", "timeout", "0"),
            ("sleep", "sleep", "0"),
            ("recipe", "cloudFormationRecipe", None),
        ]

    cloudformation_stack_mapping = [
        ("name", "stackName", None),
        ("access-key", "awsAccessKey", None),
        ("secret-key", "awsSecretKey", None),
        ("region", "awsRegion", None, region_dict),
    ]
    for map in mapping:
        cloudformation_stack_mapping.append(map)

    convert_mapping_to_xml(
        step, stack, cloudformation_stack_mapping, fail_required=True
    )


def include_exclude_patterns(xml_parent, data, yaml_prefix, xml_elem_name):
    xml_element = XML.SubElement(xml_parent, xml_elem_name)
    XML.SubElement(xml_element, "includePatterns").text = ",".join(
        data.get(yaml_prefix + "-include-patterns", [])
    )
    XML.SubElement(xml_element, "excludePatterns").text = ",".join(
        data.get(yaml_prefix + "-exclude-patterns", [])
    )


def artifactory_deployment_patterns(xml_parent, data):
    include_exclude_patterns(
        xml_parent, data, "deployment", "artifactDeploymentPatterns"
    )


def artifactory_env_vars_patterns(xml_parent, data):
    include_exclude_patterns(xml_parent, data, "env-vars", "envVarsPatterns")


def artifactory_optional_props(xml_parent, data, target):
    optional_str_props = [
        ("scopes", "scopes"),
        ("violationRecipients", "violation-recipients"),
        ("blackDuckAppName", "black-duck-app-name"),
        ("blackDuckAppVersion", "black-duck-app-version"),
        ("blackDuckReportRecipients", "black-duck-report-recipients"),
        ("blackDuckScopes", "black-duck-scopes"),
    ]

    for (xml_prop, yaml_prop) in optional_str_props:
        XML.SubElement(xml_parent, xml_prop).text = data.get(yaml_prop, "")

    common_bool_props = [
        # yaml property name, xml property name, default value
        ("deploy-artifacts", "deployArtifacts", True),
        ("discard-old-builds", "discardOldBuilds", False),
        ("discard-build-artifacts", "discardBuildArtifacts", False),
        ("publish-build-info", "deployBuildInfo", False),
        ("env-vars-include", "includeEnvVars", False),
        ("run-checks", "runChecks", False),
        ("include-publish-artifacts", "includePublishArtifacts", False),
        ("license-auto-discovery", "licenseAutoDiscovery", True),
        ("enable-issue-tracker-integration", "enableIssueTrackerIntegration", False),
        ("aggregate-build-issues", "aggregateBuildIssues", False),
        ("black-duck-run-checks", "blackDuckRunChecks", False),
        (
            "black-duck-include-published-artifacts",
            "blackDuckIncludePublishedArtifacts",
            False,
        ),
        (
            "auto-create-missing-component-requests",
            "autoCreateMissingComponentRequests",
            True,
        ),
        (
            "auto-discard-stale-component-requests",
            "autoDiscardStaleComponentRequests",
            True,
        ),
        (
            "filter-excluded-artifacts-from-build",
            "filterExcludedArtifactsFromBuild",
            False,
        ),
    ]
    convert_mapping_to_xml(xml_parent, data, common_bool_props, fail_required=True)

    if "wrappers" in target:
        wrapper_bool_props = [
            ("enable-resolve-artifacts", "enableResolveArtifacts", False),
            ("disable-license-auto-discovery", "disableLicenseAutoDiscovery", False),
            ("record-all-dependencies", "recordAllDependencies", False),
        ]
        convert_mapping_to_xml(xml_parent, data, wrapper_bool_props, fail_required=True)

    if "publishers" in target:
        publisher_bool_props = [
            ("even-if-unstable", "evenIfUnstable", False),
            ("pass-identified-downstream", "passIdentifiedDownstream", False),
            (
                "allow-promotion-of-non-staged-builds",
                "allowPromotionOfNonStagedBuilds",
                False,
            ),
        ]
        convert_mapping_to_xml(
            xml_parent, data, publisher_bool_props, fail_required=True
        )


def artifactory_common_details(details, data):
    mapping = [("name", "artifactoryName", ""), ("url", "artifactoryUrl", "")]
    convert_mapping_to_xml(details, data, mapping, fail_required=True)


def artifactory_repository(xml_parent, data, target):
    if "release" in target:
        release_mapping = [
            ("deploy-release-repo-key", "keyFromText", ""),
            ("deploy-release-repo-key", "keyFromSelect", ""),
            ("deploy-dynamic-mode", "dynamicMode", False),
        ]
        convert_mapping_to_xml(xml_parent, data, release_mapping, fail_required=True)

    if "snapshot" in target:
        snapshot_mapping = [
            ("deploy-snapshot-repo-key", "keyFromText", ""),
            ("deploy-snapshot-repo-key", "keyFromSelect", ""),
            ("deploy-dynamic-mode", "dynamicMode", False),
        ]
        convert_mapping_to_xml(xml_parent, data, snapshot_mapping, fail_required=True)


def append_git_revision_config(parent, config_def):
    params = XML.SubElement(parent, "hudson.plugins.git.GitRevisionBuildParameters")

    try:
        # If git-revision is a boolean, the get() will
        # throw an AttributeError
        combine_commits = str(config_def.get("combine-queued-commits", False)).lower()
    except AttributeError:
        combine_commits = "false"

    XML.SubElement(params, "combineQueuedCommits").text = combine_commits


def test_fairy_common(xml_element, data):
    xml_element.set("plugin", "TestFairy")
    valid_max_duration = ["10m", "60m", "300m", "1440m"]
    valid_interval = [1, 2, 5]
    valid_video_quality = ["high", "medium", "low"]

    mappings = [
        # General
        ("apikey", "apiKey", None),
        ("appfile", "appFile", None),
        ("tester-groups", "testersGroups", ""),
        ("notify-testers", "notifyTesters", True),
        ("autoupdate", "autoUpdate", True),
        # Session
        ("max-duration", "maxDuration", "10m", valid_max_duration),
        ("record-on-background", "recordOnBackground", False),
        ("data-only-wifi", "dataOnlyWifi", False),
        # Video
        ("video-enabled", "isVideoEnabled", True),
        ("screenshot-interval", "screenshotInterval", 1, valid_interval),
        ("video-quality", "videoQuality", "high", valid_video_quality),
        # Metrics
        ("cpu", "cpu", True),
        ("memory", "memory", True),
        ("logs", "logs", True),
        ("network", "network", False),
        ("phone-signal", "phoneSignal", False),
        ("wifi", "wifi", False),
        ("gps", "gps", False),
        ("battery", "battery", False),
        ("opengl", "openGl", False),
        # Advanced options
        ("advanced-options", "advancedOptions", ""),
    ]
    convert_mapping_to_xml(xml_element, data, mappings, fail_required=True)


def trigger_get_parameter_order(registry, plugin):
    logger = logging.getLogger("%s:trigger_get_parameter_order" % __name__)

    if (
        str(
            registry.jjb_config.get_plugin_config(plugin, "param_order_from_yaml", True)
        ).lower()
        == "false"
    ):
        logger.warning(
            "Using deprecated order for parameter sets in %s. It is "
            "recommended that you update your job definition instead of "
            "enabling use of the old hardcoded order",
            plugin,
        )

        # deprecated order
        return [
            "predefined-parameters",
            "git-revision",
            "property-file",
            "current-parameters",
            "node-parameters",
            "svn-revision",
            "restrict-matrix-project",
            "node-label-name",
            "node-label",
            "boolean-parameters",
        ]

    return None


def trigger_project(tconfigs, project_def, registry, param_order=None):

    plugin_ver = registry.get_plugin_version("parameterized-trigger")

    logger = logging.getLogger("%s:trigger_project" % __name__)
    pt_prefix = "hudson.plugins.parameterizedtrigger."
    if param_order:
        parameters = param_order
    else:
        parameters = project_def.keys()

    for param_type in parameters:
        param_value = project_def.get(param_type)
        if param_value is None:
            continue

        if param_type == "predefined-parameters":
            params = XML.SubElement(tconfigs, pt_prefix + "PredefinedBuildParameters")
            properties = XML.SubElement(params, "properties")
            properties.text = param_value
        elif param_type == "git-revision" and param_value:
            if "combine-queued-commits" in project_def:
                logger.warning(
                    "'combine-queued-commit' has moved to reside under "
                    "'git-revision' configuration, please update your "
                    "configs as support for this will be removed."
                )
                git_revision = {
                    "combine-queued-commits": project_def["combine-queued-commits"]
                }
            else:
                git_revision = project_def["git-revision"]
            append_git_revision_config(tconfigs, git_revision)
        elif param_type == "property-file":
            params = XML.SubElement(tconfigs, pt_prefix + "FileBuildParameters")
            property_file_mapping = [
                ("property-file", "propertiesFile", None),
                ("fail-on-missing", "failTriggerOnMissing", False),
            ]

            if plugin_ver >= "2.35.2":
                property_file_mapping.append(
                    ("property-multiline", "textParamValueOnNewLine", False)
                )

            convert_mapping_to_xml(
                params, project_def, property_file_mapping, fail_required=True
            )
            if "file-encoding" in project_def:
                XML.SubElement(params, "encoding").text = project_def["file-encoding"]
            if "use-matrix-child-files" in project_def:
                # TODO: These parameters only affect execution in
                # publishers of matrix projects; we should warn if they are
                # used in other contexts.
                use_matrix_child_files_mapping = [
                    ("use-matrix-child-files", "useMatrixChild", None),
                    ("matrix-child-combination-filter", "combinationFilter", ""),
                    ("only-exact-matrix-child-runs", "onlyExactRuns", False),
                ]
                convert_mapping_to_xml(
                    params,
                    project_def,
                    use_matrix_child_files_mapping,
                    fail_required=True,
                )
        elif param_type == "current-parameters" and param_value:
            XML.SubElement(tconfigs, pt_prefix + "CurrentBuildParameters")
        elif param_type == "node-parameters" and param_value:
            XML.SubElement(tconfigs, pt_prefix + "NodeParameters")
        elif param_type == "svn-revision" and param_value:
            param = XML.SubElement(
                tconfigs, pt_prefix + "SubversionRevisionBuildParameters"
            )
            XML.SubElement(param, "includeUpstreamParameters").text = str(
                project_def.get("include-upstream", False)
            ).lower()
        elif param_type == "restrict-matrix-project" and param_value:
            subset = XML.SubElement(
                tconfigs, pt_prefix + "matrix.MatrixSubsetBuildParameters"
            )
            XML.SubElement(subset, "filter").text = project_def[
                "restrict-matrix-project"
            ]
        elif param_type == "node-label-name" or param_type == "node-label":
            tag_name = (
                "org.jvnet.jenkins.plugins.nodelabelparameter."
                "parameterizedtrigger.NodeLabelBuildParameter"
            )
            if tconfigs.find(tag_name) is not None:
                # already processed and can only have one
                continue
            params = XML.SubElement(tconfigs, tag_name)
            name = XML.SubElement(params, "name")
            if "node-label-name" in project_def:
                name.text = project_def["node-label-name"]
            label = XML.SubElement(params, "nodeLabel")
            if "node-label" in project_def:
                label.text = project_def["node-label"]
        elif param_type == "boolean-parameters" and param_value:
            params = XML.SubElement(tconfigs, pt_prefix + "BooleanParameters")
            config_tag = XML.SubElement(params, "configs")
            param_tag_text = pt_prefix + "BooleanParameterConfig"
            params_list = param_value
            for name, value in params_list.items():
                param_tag = XML.SubElement(config_tag, param_tag_text)
                mapping = [("", "name", name), ("", "value", value or False)]
                convert_mapping_to_xml(
                    param_tag, project_def, mapping, fail_required=True
                )


def trigger_threshold(
    parent_element, element_name, threshold_name, supported_thresholds=None
):
    """Generate a resultThreshold XML element for build/join triggers"""
    element = XML.SubElement(parent_element, element_name)

    try:
        threshold = hudson_model.THRESHOLDS[threshold_name.upper()]
    except KeyError:
        if not supported_thresholds:
            supported_thresholds = hudson_model.THRESHOLDS.keys()
        raise JenkinsJobsException(
            "threshold must be one of %s" % ", ".join(supported_thresholds)
        )
    XML.SubElement(element, "name").text = threshold["name"]
    XML.SubElement(element, "ordinal").text = threshold["ordinal"]
    XML.SubElement(element, "color").text = threshold["color"]
    return element


def convert_mapping_to_xml(parent, data, mapping, fail_required=True):
    """Convert mapping to XML

    fail_required affects the last parameter of the mapping field when it's
    parameter is set to 'None'. When fail_required is True then a 'None' value
    represents a required configuration so will raise a MissingAttributeError
    if the user does not provide the configuration.

    If fail_required is False parameter is treated as optional. Logic will skip
    configuring the XML tag for the parameter. We recommend for new plugins to
    set fail_required=True and instead of optional parameters provide a default
    value for all parameters that are not required instead.

    valid_options provides a way to check if the value the user input is from a
    list of available options. When the user pass a value that is not supported
    from the list, it raise an InvalidAttributeError.

    valid_dict provides a way to set options through their key and value. If
    the user input corresponds to a key, the XML tag will use the key's value
    for its element. When the user pass a value that there are no keys for,
    it raise an InvalidAttributeError.
    """
    for elem in mapping:
        (optname, xmlname, val) = elem[:3]
        val = data.get(optname, val)

        valid_options = []
        valid_dict = {}
        if len(elem) == 4:
            if type(elem[3]) is list:
                valid_options = elem[3]
            if type(elem[3]) is dict:
                valid_dict = elem[3]

        # Use fail_required setting to allow support for optional parameters
        if val is None and fail_required is True:
            raise MissingAttributeError(optname)

        # if no value is provided then continue else leave it
        # up to the user if they want to use an empty XML tag
        if val is None and fail_required is False:
            continue

        if valid_dict:
            if val not in valid_dict:
                raise InvalidAttributeError(optname, val, valid_dict.keys())

        if valid_options:
            if val not in valid_options:
                raise InvalidAttributeError(optname, val, valid_options)

        if type(val) is bool:
            val = str(val).lower()

        if val in valid_dict:
            XML.SubElement(parent, xmlname).text = str(valid_dict[val])
        else:
            XML.SubElement(parent, xmlname).text = str(val)


def jms_messaging_common(parent, subelement, data):
    """JMS common helper function

    Pass the XML parent and the specific subelement from the builder or the
    publisher.

    data is passed to mapper helper function to map yaml fields to XML fields
    """
    namespace = XML.SubElement(parent, subelement)

    if "override-topic" in data:
        overrides = XML.SubElement(namespace, "overrides")
        XML.SubElement(overrides, "topic").text = str(data.get("override-topic", ""))

    mapping = [
        # option, xml name, default value
        ("provider-name", "providerName", ""),
        ("msg-type", "messageType", "CodeQualityChecksDone"),
        ("msg-props", "messageProperties", ""),
        ("msg-content", "messageContent", ""),
    ]
    convert_mapping_to_xml(namespace, data, mapping, fail_required=True)


def check_mutual_exclusive_data_args(data_arg_position, *args):
    mutual_exclusive_args = set(args)

    def validator(f):
        @wraps(f)
        def wrap(*args):
            actual_args = set(args[data_arg_position].keys())
            if len(actual_args & mutual_exclusive_args) > 1:
                raise JenkinsJobsException(
                    "Args: {} in {} are mutual exclusive. Please define one of it.".format(
                        mutual_exclusive_args, f.__name__
                    )
                )
            return f(*args)

        return wrap

    return validator



#
#  - Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os.path

import pytest

import jenkins_jobs


# Override fixture - do not use this mock.
@pytest.fixture(autouse=True)
def job_cache_mocked(mocker):
    pass


def test_save_on_exit(mocker):
    """
    Test that the cache is saved on normal object deletion
    """
    mocker.patch("jenkins_jobs.builder.JobCache.get_cache_dir", lambda x: "/bad/file")

    save_mock = mocker.patch("jenkins_jobs.builder.JobCache.save")
    mocker.patch("os.path.isfile", return_value=False)
    mocker.patch("jenkins_jobs.builder.JobCache._lock")
    jenkins_jobs.builder.JobCache("dummy")
    save_mock.assert_called_with()


def test_cache_file(mocker):
    """
    Test providing a cachefile.
    """
    mocker.patch("jenkins_jobs.builder.JobCache.get_cache_dir", lambda x: "/bad/file")

    test_file = os.path.abspath(__file__)
    mocker.patch("os.path.join", return_value=test_file)
    mocker.patch("yaml.safe_load")
    mocker.patch("jenkins_jobs.builder.JobCache._lock")
    jenkins_jobs.builder.JobCache("dummy").data = None



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import sys

from jenkins_jobs.errors import JenkinsJobsException
import jenkins_jobs.cli.subcommand.base as base


logger = logging.getLogger(__name__)


class UpdateSubCommand(base.JobsSubCommand):
    def parse_arg_path(self, parser):
        parser.add_argument(
            "path",
            nargs="?",
            default=sys.stdin,
            help="colon-separated list of paths to YAML files " "or directories",
        )

    def parse_arg_names(self, parser):
        parser.add_argument("names", help="name(s) of job(s)", nargs="*")

    def parse_args(self, subparser):
        update = subparser.add_parser("update")

        self.parse_option_recursive_exclude(update)

        self.parse_arg_path(update)
        self.parse_arg_names(update)

        update.add_argument(
            "--delete-old",
            action="store_true",
            dest="delete_old",
            default=False,
            help="delete obsolete jobs",
        )
        update.add_argument(
            "-p",
            "--plugin-info",
            dest="plugins_info_path",
            default=None,
            help="path to plugin info YAML file. Can be used to provide "
            "previously retrieved plugins info when connecting credentials "
            "don't have permissions to query.",
        )
        update.add_argument(
            "--workers",
            type=int,
            default=1,
            dest="n_workers",
            help="number of workers to use, 0 for autodetection and 1 "
            "for just one worker.",
        )
        update.add_argument(
            "--existing-only",
            action="store_true",
            default=False,
            dest="existing_only",
            help="update existing jobs only",
        )

        update.add_argument(
            "--enabled-only",
            action="store_true",
            default=False,
            dest="enabled_only",
            help="update enabled jobs only",
        )

        update_type = update.add_mutually_exclusive_group()
        update_type.add_argument(
            "-j",
            "--jobs-only",
            action="store_const",
            dest="update",
            const="jobs",
            help="update only jobs",
        )
        update_type.add_argument(
            "-v",
            "--views-only",
            action="store_const",
            dest="update",
            const="views",
            help="update only views",
        )

    def execute(self, options, jjb_config):
        if options.n_workers < 0:
            raise JenkinsJobsException(
                "Number of workers must be equal or greater than 0"
            )

        builder, xml_jobs, xml_views = self.make_jobs_and_views_xml(
            jjb_config, options.path, options.names
        )

        if options.enabled_only:
            # filter out jobs which are disabled
            xml_jobs_filtered = []
            for xml_job in xml_jobs:
                el = xml_job.xml.find("./disabled")
                if el is not None:
                    if el.text == "true":
                        continue
                xml_jobs_filtered.append(xml_job)

            logging.info(
                "Will only deploy enabled jobs "
                "(skipping {} disabled jobs)".format(
                    len(xml_jobs) - len(xml_jobs_filtered)
                )
            )
            xml_jobs = xml_jobs_filtered

        if options.update in {"jobs", "all"}:
            jobs, num_updated_jobs = builder.update_jobs(
                xml_jobs,
                n_workers=options.n_workers,
                existing_only=options.existing_only,
            )
            logger.info("Number of jobs updated: %d", num_updated_jobs)
        if options.update in {"views", "all"}:
            views, num_updated_views = builder.update_views(
                xml_views,
                n_workers=options.n_workers,
                existing_only=options.existing_only,
            )
            logger.info("Number of views updated: %d", num_updated_views)

        if options.delete_old:
            if options.update in {"jobs", "all"}:
                keep_jobs = [job.name for job in xml_jobs]
                n = builder.delete_old_managed_jobs(keep=keep_jobs)
                logger.info("Number of jobs deleted: %d", n)
            if options.update in {"views", "all"}:
                keep_views = [view.name for view in xml_views]
                n = builder.delete_old_managed_views(keep=keep_views)
                logger.info("Number of views deleted: %d", n)



import pytest

from jenkins_jobs.loc_loader import LocDict, LocList
from jenkins_jobs.dimensions import enum_dimensions_params, is_point_included


# Axes, params, exclude, expected resulting params.
cases = [
    (
        ["axis1"],
        {"axis1": 123},
        [],
        [{"axis1": 123}],
    ),
    (
        ["axis1"],
        {"axis1": [123, 456]},
        [],
        [
            {"axis1": 123},
            {"axis1": 456},
        ],
    ),
    (
        ["axis1", "axis2"],
        {"axis1": 123, "axis2": 456},
        [],
        [{"axis1": 123, "axis2": 456}],
    ),
    (
        ["axis1", "axis2"],
        {"axis1": [11, 22], "axis2": 456},
        [],
        [
            {"axis1": 11, "axis2": 456},
            {"axis1": 22, "axis2": 456},
        ],
    ),
    (
        ["axis1", "axis2"],
        {"axis1": [11, 22], "axis2": [33, 44, 55]},
        [],
        [
            {"axis1": 11, "axis2": 33},
            {"axis1": 11, "axis2": 44},
            {"axis1": 11, "axis2": 55},
            {"axis1": 22, "axis2": 33},
            {"axis1": 22, "axis2": 44},
            {"axis1": 22, "axis2": 55},
        ],
    ),
    (
        ["axis1", "axis2", "axis3"],
        {
            "axis1": ["axis1val1", "axis1val2"],
            "axis2": ["axis2val1", "axis2val2"],
            "axis3": ["axis3val1", "axis3val2"],
        },
        [],
        [
            {"axis1": "axis1val1", "axis2": "axis2val1", "axis3": "axis3val1"},
            {"axis1": "axis1val1", "axis2": "axis2val1", "axis3": "axis3val2"},
            {"axis1": "axis1val1", "axis2": "axis2val2", "axis3": "axis3val1"},
            {"axis1": "axis1val1", "axis2": "axis2val2", "axis3": "axis3val2"},
            {"axis1": "axis1val2", "axis2": "axis2val1", "axis3": "axis3val1"},
            {"axis1": "axis1val2", "axis2": "axis2val1", "axis3": "axis3val2"},
            {"axis1": "axis1val2", "axis2": "axis2val2", "axis3": "axis3val1"},
            {"axis1": "axis1val2", "axis2": "axis2val2", "axis3": "axis3val2"},
        ],
    ),
    # Value with parameters.
    (
        ["axis1"],
        {
            "axis1": [
                {
                    123: {
                        "param_1": "value_1",
                        "param_2": "value_2",
                    }
                },
            ]
        },
        [],
        [
            {"axis1": 123, "param_1": "value_1", "param_2": "value_2"},
        ],
    ),
    (
        ["axis1"],
        {
            "axis1": [
                {
                    "one": {
                        "param_1": "value_1_one",
                        "param_2": "value_2_one",
                    }
                },
                {
                    "two": {
                        "param_1": "value_1_two",
                        "param_2": "value_2_two",
                    }
                },
            ]
        },
        [],
        [
            {"axis1": "one", "param_1": "value_1_one", "param_2": "value_2_one"},
            {"axis1": "two", "param_1": "value_1_two", "param_2": "value_2_two"},
        ],
    ),
    (
        ["axis1"],
        {
            "axis1": [
                {
                    "one": {
                        "param_1": "value_1_one",
                        "param_2": "value_2_one",
                    }
                },
                "two",
            ]
        },
        [],
        [
            {"axis1": "one", "param_1": "value_1_one", "param_2": "value_2_one"},
            {"axis1": "two"},
        ],
    ),
    # Axis value received from another axis parameter.
    (
        ["axis1", "axis2"],
        {
            "axis1": [
                {"one": {"axis2": "axis2_value"}},
            ],
        },
        [],
        [{"axis1": "one", "axis2": "axis2_value"}],
    ),
    # With excludes.
    (
        ["axis1", "axis2", "axis3"],
        {
            "axis1": "axis1val",
            "axis2": ["axis2val1", "axis2val2"],
            "axis3": ["axis3val1", "axis3val2"],
        },
        [
            {  # First exclude.
                "axis2": "axis2val1",
                "axis3": "axis3val2",
            },
            {  # Second exclude.
                "axis3": "axis3val1",
            },
        ],
        [
            # Excluded by second: {"axis1": "axis1val", "axis2": "axis2val1", "axis3": "axis3val1"},
            # Excluded by first: {"axis1": "axis1val", "axis2": "axis2val1", "axis3": "axis3val2"},
            # Excluded by second: {"axis1": "axis1val", "axis2": "axis2val2", "axis3": "axis3val1"},
            {"axis1": "axis1val", "axis2": "axis2val2", "axis3": "axis3val2"},
        ],
    ),
    (
        ["axis1", "axis2", "axis3"],
        {
            "axis1": ["axis1val1", "axis1val2"],
            "axis2": ["axis2val1", "axis2val2"],
            "axis3": ["axis3val1", "axis3val2"],
        },
        [
            {  # First exclude.
                "axis1": "axis1val1",
                "axis2": "axis2val1",
                "axis3": "axis3val2",
            },
            {  # Second exclude.
                "axis2": "axis2val2",
                "axis3": "axis3val1",
            },
        ],
        [
            {"axis1": "axis1val1", "axis2": "axis2val1", "axis3": "axis3val1"},
            # Excluded by first: {"axis1": "axis1val1", "axis2": "axis2val1", "axis3": "axis3val2"},
            # Excluded by second: {"axis1": "axis1val1", "axis2": "axis2val2", "axis3": "axis3val1"},
            {"axis1": "axis1val1", "axis2": "axis2val2", "axis3": "axis3val2"},
            {"axis1": "axis1val2", "axis2": "axis2val1", "axis3": "axis3val1"},
            {"axis1": "axis1val2", "axis2": "axis2val1", "axis3": "axis3val2"},
            # Excluded by second: {"axis1": "axis1val2", "axis2": "axis2val2", "axis3": "axis3val1"},
            {"axis1": "axis1val2", "axis2": "axis2val2", "axis3": "axis3val2"},
        ],
    ),
]


def wrap_with_location(value):
    if type(value) is dict:
        return LocDict({key: wrap_with_location(value) for key, value in value.items()})
    if type(value) is list:
        return LocList([wrap_with_location(item) for item in value])
    return value


@pytest.mark.parametrize("axes,params,exclude,expected_dimension_params", cases)
def test_dimensions(axes, params, exclude, expected_dimension_params):
    dimension_params = [
        p
        for p in enum_dimensions_params(axes, wrap_with_location(params), defaults={})
        if is_point_included(LocList(exclude), p)
    ]
    assert dimension_params == expected_dimension_params



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The Notifications module allows you to configure Jenkins to notify
other applications about various build phases.  It requires the
Jenkins notification plugin.

**Component**: notifications
  :Macro: notification
  :Entry Point: jenkins_jobs.notifications

"""

import xml.etree.ElementTree as XML

import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers


def http_endpoint(registry, xml_parent, data):
    """yaml: http
    Defines an HTTP notification endpoint.

    Requires the Jenkins :jenkins-plugins:`Notification Plugin
    <notification>`.

    :arg str format: notification payload format, JSON (default) or XML
    :arg str event: job events that trigger notifications: started,
        completed, finalized or all (default)
    :arg str url: URL of the endpoint
    :arg int timeout: Timeout in milliseconds for sending notification
        request (30 seconds by default)
    :arg int retries: Nr of times to retry sending notification in case
        sending notification fails. (0 by default)
    :arg int log: Number lines of log messages to send (0 by default).
        Use -1 for all (use with caution).

    Example:

    .. literalinclude:: \
    /../../tests/notifications/fixtures/http-endpoint002.yaml
       :language: yaml

    """
    endpoint_element = XML.SubElement(
        xml_parent, "com.tikal.hudson.plugins.notification." "Endpoint"
    )
    supported_formats = ["JSON", "XML"]
    supported_events = ["started", "completed", "finalized", "all"]
    fmt = data.get("format", "JSON").upper()
    event = data.get("event", "all").lower()
    mapping = [
        ("", "format", fmt, supported_formats),
        ("", "protocol", "HTTP"),
        ("", "event", event, supported_events),
        ("timeout", "timeout", 30000),
        ("retries", "retries", 0),
        ("url", "url", None),
        ("log", "loglines", 0),
    ]
    helpers.convert_mapping_to_xml(endpoint_element, data, mapping, fail_required=True)


class Notifications(jenkins_jobs.modules.base.Base):
    sequence = 22

    component_type = "notification"
    component_list_type = "notifications"

    def gen_xml(self, xml_parent, data):
        properties = xml_parent.find("properties")
        if properties is None:
            properties = XML.SubElement(xml_parent, "properties")

        notifications = data.get("notifications", [])
        if notifications:
            notify_element = XML.SubElement(
                properties,
                "com.tikal.hudson.plugins."
                "notification."
                "HudsonNotificationProperty",
            )
            endpoints_element = XML.SubElement(notify_element, "endpoints")

            self.dispatch_component_list(
                "notification", notifications, endpoints_element
            )



import pytest

from jenkins_jobs.cli import entry


def test_with_empty_args(mocker):
    """
    User passes no args, should fail with SystemExit
    """
    with pytest.raises(SystemExit):
        entry.JenkinsJobs([])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from functools import partial
from itertools import filterfalse
from jinja2 import StrictUndefined

from .errors import Context, JenkinsJobsException
from .formatter import CustomFormatter, enum_str_format_required_params
from .loc_loader import LocDict, LocString, LocList
from .yaml_objects import (
    J2String,
    J2Yaml,
    YamlInclude,
    YamlListJoin,
    IncludeJinja2,
    IncludeRawExpand,
    IncludeRawVerbatim,
)


def expand_dict(expander, obj, params, key_pos, value_pos):
    result = LocDict(pos=obj.pos)
    for key, value in obj.items():
        expanded_key = expander.expand(key, params, None)
        expanded_value = expander.expand(
            value, params, obj.key_pos.get(key), obj.value_pos.get(key)
        )
        result.set_item(
            expanded_key, expanded_value, obj.key_pos.get(key), obj.value_pos.get(key)
        )
    return result


def expand_list(expander, obj, params, key_pos, value_pos):
    items = [
        expander.expand(item, params, None, obj.value_pos[idx])
        for idx, item in enumerate(obj)
    ]
    value_pos = [obj.value_pos[idx] for idx, _ in enumerate(obj)]
    return LocList(items, obj.pos, value_pos)


def expand_tuple(expander, obj, params, key_pos, value_pos):
    return tuple(expander.expand(item, params, None) for item in obj)


class StrExpander:
    def __init__(self, allow_empty_variables):
        self._formatter = CustomFormatter(allow_empty_variables)

    def __call__(self, obj, params, key_pos, value_pos):
        try:
            return self._formatter.format(str(obj), **params)
        except JenkinsJobsException as x:
            lines = str(obj).splitlines()
            start_ofs = value_pos.body.index(lines[0])
            pre_pad = value_pos.body[:start_ofs]
            # Shift position to reflect template position inside yaml file:
            if "\n" in pre_pad:
                pos = value_pos.with_offset(line_ofs=1)
            else:
                pos = value_pos.with_offset(column_ofs=start_ofs)
            pos = pos.with_contents_start()
            raise x.with_pos(pos)


def call_expand(expander, obj, params, key_pos, value_pos):
    return obj.expand(expander, params)


def dont_expand(obj, params, key_pos, value_pos):
    return obj


def dont_expand_yaml_object(expander, obj, params, key_pos, value_pos):
    return obj


yaml_classes_list = [
    J2String,
    J2Yaml,
    YamlInclude,
    YamlListJoin,
    IncludeJinja2,
    IncludeRawExpand,
    IncludeRawVerbatim,
]

deprecated_yaml_tags = [
    ("!include", YamlInclude),
    ("!include-raw", IncludeRawExpand),
    ("!include-raw:", IncludeRawExpand),
    ("!include-raw-escape", IncludeRawVerbatim),
    ("!include-raw-escape:", IncludeRawVerbatim),
]


# Expand strings and yaml objects.
class Expander:
    def __init__(self, config=None):
        if config:
            allow_empty_variables = config.yamlparser["allow_empty_variables"]
        else:
            allow_empty_variables = False
        _yaml_object_expanders = {
            cls: partial(call_expand, self) for cls in yaml_classes_list
        }
        self.expanders = {
            dict: partial(expand_dict, self),
            LocDict: partial(expand_dict, self),
            list: partial(expand_list, self),
            LocList: partial(expand_list, self),
            tuple: partial(expand_tuple, self),
            str: StrExpander(allow_empty_variables),
            LocString: StrExpander(allow_empty_variables),
            bool: dont_expand,
            int: dont_expand,
            float: dont_expand,
            type(None): dont_expand,
            **_yaml_object_expanders,
        }

    def expand(self, obj, params, key_pos=None, value_pos=None):
        t = type(obj)
        try:
            expander = self.expanders[t]
        except KeyError:
            raise JenkinsJobsException(
                f"Do not know how to expand type: {t!r}", pos=value_pos
            )
        return expander(obj, params, key_pos, value_pos)


# Expand only yaml objects.
class YamlObjectsExpander(Expander):
    def __init__(self, config=None):
        super().__init__(config)
        self.expanders.update(
            {
                str: dont_expand,
                LocString: dont_expand,
                IncludeRawVerbatim: dont_expand,
            }
        )


# Expand only string parameters.
class StringsOnlyExpander(Expander):
    def __init__(self, config):
        super().__init__(config)
        _yaml_object_expanders = {
            cls: partial(dont_expand_yaml_object, self) for cls in yaml_classes_list
        }
        self.expanders.update(_yaml_object_expanders)


def call_required_params(obj, pos):
    yield from obj.required_params


def enum_dict_params(obj, pos):
    for key, value in obj.items():
        yield from enum_required_params(key, obj.key_pos.get(key))
        yield from enum_required_params(value, obj.value_pos.get(key))


def enum_seq_params(obj, pos):
    for idx, value in enumerate(obj):
        yield from enum_required_params(value, pos=None)


def enum_loc_list_params(obj, pos):
    for idx, value in enumerate(obj):
        yield from enum_required_params(value, obj.value_pos[idx])


def no_parameters(obj, pos):
    return []


yaml_classes_enumers = {cls: call_required_params for cls in yaml_classes_list}

param_enumers = {
    str: enum_str_format_required_params,
    LocString: enum_str_format_required_params,
    dict: enum_dict_params,
    LocDict: enum_dict_params,
    list: enum_seq_params,
    LocList: enum_loc_list_params,
    tuple: enum_seq_params,
    bool: no_parameters,
    int: no_parameters,
    float: no_parameters,
    type(None): no_parameters,
    **yaml_classes_enumers,
}

# Do not expand these.
disable_expand_for = {"template-name"}


def enum_required_params(obj, pos):
    t = type(obj)
    try:
        enumer = param_enumers[t]
    except KeyError:
        raise JenkinsJobsException(
            f"Do not know how to enumerate required parameters for type: {t!r}",
            pos=pos,
        )
    return enumer(obj, pos)


def expand_parameters(expander, param_dict):
    expanded_params = LocDict()
    deps = {}  # Variable name -> variable pos.

    def deps_context():
        return [Context(f"Used by {n}", vp) for n, (kp, vp) in deps.items()]

    def expand(name):
        try:
            value = expanded_params[name]
            key_pos = expanded_params.key_pos.get(name)
            value_pos = expanded_params.value_pos.get(name)
            return (value, key_pos, value_pos)
        except KeyError:
            pass
        try:
            format = param_dict[name]
        except KeyError:
            return (StrictUndefined(name=name), None, None)
        key_pos = param_dict.key_pos.get(name)
        value_pos = param_dict.value_pos.get(name)
        if name in deps:
            expand_ctx = Context(f"While expanding {name!r}", key_pos)
            raise JenkinsJobsException(
                f"Recursive parameters usage: {' <- '.join(deps)}",
                pos=value_pos,
                ctx=[*deps_context(), expand_ctx],
            )
        if name in disable_expand_for:
            value = format
        else:
            required_params = list(enum_required_params(format, value_pos))
            deps[name] = (key_pos, value_pos)
            try:
                params = LocDict.merge(expanded_params)
                for n in required_params:
                    v, kp, vp = expand(n)
                    params.set_item(n, v, kp, vp)
            finally:
                deps.popitem()
            try:
                value = expander.expand(format, params, key_pos, value_pos)
            except JenkinsJobsException as x:
                raise x.with_context(
                    f"While expanding parameter {name!r}",
                    pos=key_pos,
                    ctx=deps_context(),
                )
        expanded_params.set_item(name, value, key_pos, value_pos)
        return (value, key_pos, value_pos)

    expand("name")  # expand 'name' parameter first
    for name in filterfalse(lambda x: x == "name", param_dict):
        expand(name)
    return expanded_params



# Copyright 2016 Hewlett Packard Enterprise
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from pathlib import Path

import pytest

from jenkins_jobs.config import JJBConfig
from jenkins_jobs.errors import JenkinsJobsException
from jenkins_jobs.xml_config import XmlJobGenerator, XmlViewGenerator
from jenkins_jobs.roots import Roots
from jenkins_jobs.loader import load_files


fixtures_dir = Path(__file__).parent / "exceptions"


# Override jjb_config and plugins_info so that scenarios won't be used.
@pytest.fixture
def jjb_config():
    config = JJBConfig()
    config.validate()
    return config


@pytest.fixture
def plugins_info():
    return None


@pytest.fixture
def parser(jjb_config, registry):
    def parse(fname):
        roots = Roots(jjb_config)
        load_files(jjb_config, roots, [fixtures_dir / fname])
        registry.set_macros(roots.macros)
        return roots

    return parse


def test_invalid_project(parser, registry):
    roots = parser("invalid_project.yaml")
    jobs = roots.generate_jobs()
    generator = XmlJobGenerator(registry)

    with pytest.raises(JenkinsJobsException) as excinfo:
        generator.generateXML(jobs)
    assert "Unrecognized project-type:" in str(excinfo.value)


def test_invalid_view(parser, registry):
    roots = parser("invalid_view.yaml")
    views = roots.generate_views()
    generator = XmlViewGenerator(registry)

    with pytest.raises(JenkinsJobsException) as excinfo:
        generator.generateXML(views)
    assert "Unrecognized view-type:" in str(excinfo.value)


def test_template_params(parser, registry):
    roots = parser("failure_formatting_component.yaml")
    jobs = roots.generate_jobs()
    generator = XmlJobGenerator(registry)

    with pytest.raises(Exception) as excinfo:
        generator.generateXML(jobs)
    message = "While formatting string '{branches}': Missing parameter: 'branches'"
    assert str(excinfo.value) == message


def test_missing_j2_param(parser, registry):
    roots = parser("missing_j2_parameter.yaml")
    jobs = roots.generate_jobs()
    generator = XmlJobGenerator(registry)

    with pytest.raises(Exception) as excinfo:
        generator.generateXML(jobs)
    message = "'branches' is undefined"
    assert str(excinfo.value) == message


def test_missing_include_j2_param(parser, registry):
    roots = parser("missing_include_j2_parameter.yaml")
    jobs = roots.generate_jobs()
    generator = XmlJobGenerator(registry)

    with pytest.raises(Exception) as excinfo:
        generator.generateXML(jobs)
    message = "'branches' is undefined"
    assert str(excinfo.value) == message



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import functools
import io
import logging
import warnings
from functools import partial

from .errors import JenkinsJobsException
from .loc_loader import LocLoader
from .yaml_objects import BaseYamlObject
from .expander import YamlObjectsExpander, deprecated_yaml_tags, yaml_classes_list
from .roots import root_adders

logger = logging.getLogger(__name__)


class Loader(LocLoader):
    @classmethod
    def empty(cls, jjb_config):
        return cls(io.StringIO(), jjb_config)

    def __init__(
        self, stream, jjb_config, source_path=None, source_dir=None, anchors=None
    ):
        super().__init__(stream, source_path)
        self.jjb_config = jjb_config
        self.source_path = source_path
        self.source_dir = source_dir
        self._retain_anchors = jjb_config.yamlparser["retain_anchors"]
        if anchors:
            # Override default set by super class.
            self.anchors = anchors

    # Override the default composer to skip resetting the anchors at the
    # end of the current document.
    def compose_document(self):
        # Drop the DOCUMENT-START event.
        self.get_event()
        # Compose the root node.
        node = self.compose_node(None, None)
        # Drop the DOCUMENT-END event.
        self.get_event()
        return node

    def _with_stream(self, stream, source_path, source_dir):
        return Loader(stream, self.jjb_config, source_path, source_dir, self.anchors)

    def load_fp(self, fp):
        return self.load(fp)

    # 1024 cache size seems like a reasonable balance between memory requirement and performance gain
    @functools.lru_cache(maxsize=1024)
    def load_path(self, path):
        # Caching parsed file outputs is safe even with _retain_anchors set to True because:
        # PyYAML does not allow updating anchor values, it is considered as anchor duplication
        # So we can safely cache a parsed YAML for a file containing an alias since the alias can be defined
        # only once and must be defined before use. The alias value will remain same irrespective of the number
        # times a file is parsed
        return self.load(path.read_text(), source_path=path, source_dir=path.parent)

    def load(self, stream, source_path=None, source_dir=None):
        loader = self._with_stream(stream, source_path, source_dir)
        try:
            return loader.get_single_data()
        finally:
            loader.dispose()
            if self._retain_anchors:
                self.anchors.update(loader.anchors)


def load_deprecated_yaml(tag, cls, loader, node):
    warnings.warn(
        f"Tag {tag!r} is deprecated, switch to using {cls.yaml_tag!r}",
        UserWarning,
    )
    return cls.from_yaml(loader, node)


for cls in yaml_classes_list:
    Loader.add_constructor(cls.yaml_tag, cls.from_yaml)

for tag, cls in deprecated_yaml_tags:
    Loader.add_constructor(tag, partial(load_deprecated_yaml, tag, cls))


def is_stdin(path):
    return hasattr(path, "read")


def enum_expanded_paths(path_list):
    visited_set = set()

    def real(path):
        real_path = path.resolve()
        if real_path in visited_set:
            logger.warning(
                "File '%s' is already added as '%s'; ignoring reference to avoid"
                " duplicating YAML definitions.",
                path,
                real_path,
            )
        else:
            yield real_path
            visited_set.add(real_path)

    for path in path_list:
        if is_stdin(path):
            yield path
        elif path.is_dir():
            for p in path.iterdir():
                if p.suffix in {".yml", ".yaml"}:
                    yield from real(p)
        else:
            yield from real(path)


def load_files(config, roots, path_list):
    expander = YamlObjectsExpander(config)
    loader = Loader.empty(config)
    for path in enum_expanded_paths(path_list):
        if is_stdin(path):
            data = loader.load_fp(path)
        else:
            data = loader.load_path(path)
        if data is None:
            continue
        if not isinstance(data, list):
            raise JenkinsJobsException(
                f"The topmost collection must be a list, but is: {data}",
                pos=data.pos,
            )
        for idx, item in enumerate(data):
            if not isinstance(item, dict):
                raise JenkinsJobsException(
                    f"Topmost list should contain single-item dict,"
                    f" not a {type(item)}. Missing indent?",
                    pos=data.value_pos[idx],
                )
            if len(item) != 1:
                raise JenkinsJobsException(
                    f"Topmost dict should be single-item,"
                    f" but have keys {list(item.keys())}. Missing indent?",
                    pos=item.pos,
                )
            kind, contents = next(iter(item.items()))
            if kind.startswith("_"):
                continue
            if isinstance(contents, BaseYamlObject):
                contents = contents.expand(expander, params={})
            try:
                adder = root_adders[kind]
            except KeyError:
                raise JenkinsJobsException(
                    f"Unknown topmost element type : {kind!r};"
                    f" known are: {','.join(root_adders)}.",
                    pos=item.pos,
                )
            adder(config, roots, contents, item.pos)



# Copyright 2019 Openstack Foundation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The view sectioned module handles creating Jenkins Sectioned views.

To create a sectioned view specify ``sectioned`` in the ``view-type`` attribute
to the :ref:`view_sectioned` definition.

:View Parameters:
    * **name** (`str`): The name of the view.
    * **view-type** (`str`): The type of view.
    * **description** (`str`): A description of the view. (default '')
    * **filter-executors** (`bool`): Show only executors that can
      execute the included views. (default false)
    * **filter-queue** (`bool`): Show only included jobs in builder
      queue. (default false)
    * **sections** (`list`): The views to show in sections.

Example:

    .. literalinclude::
        /../../tests/views/fixtures/view_sectioned.yaml

"""

import xml.etree.ElementTree as XML
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers
from jenkins_jobs.modules import view_list


class Sectioned(jenkins_jobs.modules.base.Base):
    def root_xml(self, data):
        root = XML.Element("hudson.plugins.sectioned__view.SectionedView")

        mapping = [
            ("name", "name", None),
            ("description", "description", ""),
            ("filter-executors", "filterExecutors", False),
            ("filter-queue", "filterQueue", False),
        ]
        helpers.convert_mapping_to_xml(root, data, mapping, fail_required=True)

        XML.SubElement(root, "properties", {"class": "hudson.model.View$PropertyList"})

        s_xml = XML.SubElement(root, "sections")
        sections = data.get("sections", [])

        for section in sections:
            if "view-type" not in section:
                raise KeyError("This sectioned view is missing a view-type.", section)
            if section["view-type"] == "list":
                project = view_list.List(self.registry)
            elif section["view-type"] == "text":
                project = SectionedText(self.registry)
            else:
                raise ValueError(
                    "Nested view-type %s is not known. Supported types: list, text."
                    % section["view-type"]
                )
            xml_project = project.root_xml(section, sectioned=True)
            s_xml.append(xml_project)

        return root


class SectionedText(jenkins_jobs.modules.base.Base):
    def root_xml(self, data, sectioned=False):
        assert sectioned

        root = XML.Element("hudson.plugins.sectioned__view.TextSection")

        # these are in the XML and look like from view_list,
        # but the UI has no controls for them
        jn_xml = XML.SubElement(root, "jobNames")
        XML.SubElement(
            jn_xml, "comparator", {"class": "hudson.util.CaseInsensitiveComparator"}
        )
        XML.SubElement(root, "jobFilters")

        mapping = [
            ("name", "name", ""),
            ("width", "width", "FULL", ["FULL", "HALF", "THIRD", "TWO_THIRDS"]),
            ("alignment", "alignment", "CENTER", ["CENTER", "LEFT", "RIGHT"]),
            ("text", "text", None),
            ("style", "style", "NONE", ["NONE", "NOTE", "INFO", "WARNING", "TIP"]),
        ]
        helpers.convert_mapping_to_xml(root, data, mapping, fail_required=True)

        return root



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from jenkins_jobs.builder import JenkinsManager
from jenkins_jobs.errors import JenkinsJobsException
import jenkins_jobs.cli.subcommand.base as base


class DeleteSubCommand(base.JobsSubCommand):
    def parse_args(self, subparser):
        delete = subparser.add_parser("delete")

        self.parse_option_recursive_exclude(delete)

        delete.add_argument("name", help="name of job", nargs="+")
        delete.add_argument(
            "-p",
            "--path",
            default=None,
            help="colon-separated list of paths to YAML files " "or directories",
        )
        delete.add_argument(
            "-j",
            "--jobs-only",
            action="store_true",
            dest="del_jobs",
            default=False,
            help="delete only jobs",
        )
        delete.add_argument(
            "-v",
            "--views-only",
            action="store_true",
            dest="del_views",
            default=False,
            help="delete only views",
        )

    def execute(self, options, jjb_config):
        builder = JenkinsManager(jjb_config)

        if options.del_jobs and options.del_views:
            raise JenkinsJobsException(
                '"--views-only" and "--jobs-only" cannot be used together.'
            )

        if options.path:
            roots = self.load_roots(jjb_config, options.path)
            jobs = base.filter_matching(roots.generate_jobs(), options.name)
            views = base.filter_matching(roots.generate_views(), options.name)
            job_names = [j.name for j in jobs]
            view_names = [v.name for v in views]
        else:
            job_names = options.name
            view_names = options.name

        if options.del_jobs:
            builder.delete_jobs(job_names)
        elif options.del_views:
            builder.delete_views(view_names)
        else:
            builder.delete_jobs(job_names)
            builder.delete_views(view_names)
